from .result_reader_creator import ResultReaderCreator
from .result_reader_creator import ResultReaderType

from .result_reader import ResultReader
from .result_reader_copier import ResultReaderCopier
from .result_reader_query import ResultReaderQuery
from .result_writer import ResultWriter



import os
import pytest
import numpy as np
import pandas as pd

import System

from mikeio1d.custom_exceptions import NoDataForQuery, InvalidQuantity
from mikeio1d.res1d import Res1D, mike1d_quantities, QueryDataReach, QueryDataNode
from mikeio1d.dotnet import to_numpy, from_dotnet_datetime, to_dotnet_datetime


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # Original file name was Exam6Base.res1d
    return os.path.join(test_folder_path, "testdata", "Network.res1d")


@pytest.fixture(params=[True, False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_file_does_not_exist():
    with pytest.raises(FileExistsError):
        assert Res1D("tests/testdata/not_a_file.res1d")


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 110
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_mike1d_quantities():
    quantities = mike1d_quantities()
    assert "WaterLevel" in quantities


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 2


def test_repr(test_file):
    res1d = test_file
    res1d_repr = res1d.__repr__()
    res1d_repr_ref = (
        "<mikeio1d.Res1D>\n"
        + "Start time: 1994-08-07 16:35:00\n"
        + "End time: 1994-08-07 18:35:00\n"
        + "# Timesteps: 110\n"
        + "# Catchments: 0\n"
        + "# Nodes: 119\n"
        + "# Reaches: 118\n"
        + "# Globals: 0\n"
        + "0 - WaterLevel <m>\n"
        + "1 - Discharge <m^3/s>"
    )
    assert res1d_repr == res1d_repr_ref


def test_data_item_dicts(test_file):
    res1d = test_file
    assert len(res1d.catchments) == 0
    assert len(res1d.nodes) == 119
    assert len(res1d.reaches) == 118
    assert len(res1d.structures) == 2
    assert len(res1d.global_data) == 0


@pytest.mark.parametrize(
    "query,expected",
    [
        (QueryDataReach("WaterLevel", "104l1", 34.4131), True),
        (QueryDataReach("WaterLevel", "104l1", 42424242), False),
        (QueryDataReach("WaterLevel", "wrong_reach_name", 34.4131), False),
    ],
)
def test_valid_reach_data_queries(test_file, query, expected):
    res1d = test_file

    with pytest.raises(InvalidQuantity):
        invalid_query = QueryDataReach("InvalidQuantity", "104l1", 34.4131)
        assert res1d.read(invalid_query)

    if expected:
        res1d.read(query)
    else:
        with pytest.raises(NoDataForQuery):
            assert res1d.read(query)


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataReach("WaterLevel", "104l1", 34.4131), 197.046),
        (QueryDataReach("WaterLevel", "9l1", 10), 195.165),
        (QueryDataReach("Discharge", "100l1", 23.8414), 0.1),
        (QueryDataReach("Discharge", "9l1", 5), 0.761),
    ],
)
def test_read_reach_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


@pytest.mark.parametrize(
    "quantity,reach_id,chainage,expected_max",
    [
        ("WaterLevel", "104l1", 34.4131, 197.046),
        ("WaterLevel", "9l1", 10, 195.165),
        ("Discharge", "100l1", 23.8414, 0.1),
        ("Discharge", "9l1", 5, 0.761),
    ],
)
def test_read_reach(test_file, quantity, reach_id, chainage, expected_max):
    data = test_file.query.GetReachValues(reach_id, chainage, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


@pytest.mark.parametrize(
    "quantity,node_id,expected_max",
    [("WaterLevel", "1", 195.669), ("WaterLevel", "2", 195.823)],
)
def test_read_node(test_file, quantity, node_id, expected_max):
    data = test_file.query.GetNodeValues(node_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


def test_time_index(test_file):
    assert len(test_file.time_index) == 110


def test_start_time(test_file):
    assert test_file.start_time == test_file.time_index.min()


def test_from_dotnet_datetime_preserves_millisecond_precision():
    dtstr = "2021-01-01 00:00:00.123"

    dotnet_dt = System.DateTime.Parse(dtstr)
    assert dotnet_dt.Millisecond == 123
    py_dt = from_dotnet_datetime(dotnet_dt)
    # python datetime doesn't have a millisecond property so we use microsecond
    assert py_dt.microsecond == 123000


# Write the same test as above, but in the reverse direction
def test_to_dotnet_datetime_preserves_millisecond_precision():
    dtstr = "2021-01-01 00:00:00.123"

    py_dt = pd.to_datetime(dtstr)
    assert py_dt.microsecond == 123000
    dotnet_dt = to_dotnet_datetime(py_dt)
    assert dotnet_dt.Millisecond == 123


def test_time_index_microseconds(test_file):
    df = test_file.read()
    assert df.index.microsecond.unique().size > 1


def test_get_node_values(test_file):
    values = test_file.get_node_values("1", "WaterLevel")
    assert len(values) == 110


def test_get_reach_values(test_file):
    values = test_file.get_reach_values("9l1", 5, "WaterLevel")
    time_series = pd.Series(values, index=test_file.time_index)
    assert len(values) == 110
    assert len(time_series.index) == 110
    # Just try to call the methods
    test_file.get_reach_end_values("9l1", "WaterLevel")
    test_file.get_reach_start_values("9l1", "WaterLevel")
    test_file.get_reach_sum_values("9l1", "WaterLevel")


def test_get_reach_value(test_file):
    value = test_file.get_reach_value("9l1", 5, "WaterLevel", test_file.start_time)
    assert value > 0


def test_dotnet_methods(test_file):
    res1d = test_file
    # Just try to access the properties and methods in .net
    res1d.data.ResultSpecs
    res1d.data.Nodes
    res1d.query.GetNodeValues("1", "WaterLevel")
    res1d.query.GetReachValue(
        "9l1", 5, "WaterLevel", res1d.data.StartTime
    )  # must be dotnet datetime
    res1d.query.GetReachValues("9l1", 5, "WaterLevel")
    res1d.query.GetReachEndValues("9l1", "WaterLevel")  # avoid specifying chainage
    res1d.query.GetReachStartValues("9l1", "WaterLevel")  # avoid specifying chainage
    res1d.query.GetReachSumValues(
        "9l1", "WaterLevel"
    )  # useful for summing volume in reach (all grid points)


def test_res1d_filter(test_file_path, helpers):
    nodes = ["1", "2"]
    reaches = ["9l1"]
    res1d = Res1D(test_file_path, nodes=nodes, reaches=reaches)

    df_9l1 = res1d.read(QueryDataReach("WaterLevel", "9l1", 10))
    df_1 = res1d.read(QueryDataNode("WaterLevel", "1"))
    df_2 = res1d.read(QueryDataNode("WaterLevel", "2"))

    res1d_full = Res1D(test_file_path)
    df_full = res1d_full.read()

    helpers.assert_shared_columns_equal(df_full, df_9l1)
    helpers.assert_shared_columns_equal(df_full, df_1)
    helpers.assert_shared_columns_equal(df_full, df_2)

    # Currently Mike1D raises NullReferenceException when requesting location not included by filter
    # This should be fixed in Mike1D to raise more meaningful Mike1DException
    # with pytest.raises(Exception):
    #     assert res1d.read(QueryDataReach("WaterLevel", "100l1", 10))
    #     assert res1d.read(QueryDataNode("WaterLevel", "3"))


def test_res1d_filter_readall(test_file_path, helpers):
    # Make sure read all can be used with filters
    nodes = ["1", "2"]
    reaches = ["9l1"]
    res1d = Res1D(test_file_path, nodes=nodes, reaches=reaches)
    df = res1d.read()

    res1d_full = Res1D(test_file_path)
    df_full = res1d_full.read()

    helpers.assert_shared_columns_equal(df_full, df)


def test_res1d_filter_using_flow_split(flow_split_file_path, helpers):
    res1d_full = Res1D(flow_split_file_path)
    df_full = res1d_full.read()

    for node in res1d_full.nodes:
        res1d = Res1D(flow_split_file_path, nodes=[str(node)])
        df = res1d.read()
        helpers.assert_shared_columns_equal(df_full, df)

    for reach in res1d_full.reaches:
        res1d = Res1D(flow_split_file_path, reaches=[str(reach)])
        df = res1d.read()
        helpers.assert_shared_columns_equal(df_full, df)


def test_node_attributes(test_file):
    res1d = test_file
    nodes = res1d.nodes

    nodes.n_1.WaterLevel.add()
    nodes.n_2.WaterLevel.add()
    df = res1d.read()

    actual_max = round(df["WaterLevel:1"].max(), 3)
    assert pytest.approx(actual_max) == 195.669

    actual_max = round(df["WaterLevel:2"].max(), 3)
    assert pytest.approx(actual_max) == 195.823


def test_reach_attributes(test_file):
    res1d = test_file
    reaches = res1d.reaches

    reaches.r_104l1.m_34_4131.WaterLevel.add()

    reaches.r_9l1.m_10.WaterLevel.add()

    reaches.r_100l1.m_23_8414.Discharge.add()

    reaches.r_9l1.m_5.Discharge.add()

    df = res1d.read()

    actual_max = round(df["WaterLevel:104l1:34.4131"].max(), 3)
    assert pytest.approx(actual_max) == 197.046

    actual_max = round(df["WaterLevel:9l1:10"].max(), 3)
    assert pytest.approx(actual_max) == 195.165

    actual_max = round(df["Discharge:100l1:23.8414"].max(), 3)
    assert pytest.approx(actual_max) == 0.1

    actual_max = round(df["Discharge:9l1:5"].max(), 3)
    assert pytest.approx(actual_max) == 0.761


def test_structure_reach_attributes(test_file):
    res1d = test_file
    structures = res1d.structures

    structures.s_119w1.Discharge.add()
    structures.s_115p1.Discharge.add()

    df = res1d.read()

    max_discharge = round(df.max().max(), 3)
    assert pytest.approx(max_discharge) == 1.491


def test_structure_reach_static_attributes_exist(res1d_network, res1d_river_network):
    for network in [res1d_network, res1d_river_network]:
        for _, structure in network.result_network.structures.items():
            assert hasattr(structure, "id")
            assert hasattr(structure, "type")
            assert hasattr(structure, "chainage")


def test_structure_reach_static_attributes(res1d_network):
    structures = res1d_network.result_network.structures

    assert structures.s_119w1.type == "Weir"
    assert structures.s_119w1.id == "119w1"
    assert structures.s_119w1.chainage == pytest.approx(0.5)

    assert structures.s_115p1.type == "Pump"
    assert structures.s_115p1.id == "115p1"
    assert structures.s_115p1.chainage == pytest.approx(41.21402714094492)


def test_structure_reach_maintains_backweards_compatibility(res1d_network):
    structures = res1d_network.result_network.structures

    with pytest.warns(UserWarning):
        assert structures.s_119w1.structure_id == structures.s_119w1.id


def test_nodes_dict_access_maintains_backwards_compatibility(res1d_network):
    with pytest.warns(UserWarning):
        node = res1d_network.result_network.nodes["1"]
        assert node.GroundLevel == pytest.approx(197.07000732421875)
        assert node.BottomLevel == pytest.approx(195.0500030517578)
        assert node.XCoordinate == pytest.approx(-687934.6000976562)


def test_node_node_property_maintains_backwards_compatibility(res1d_network):
    node = res1d_network.result_network.nodes.n_1
    assert node.node.GroundLevel == pytest.approx(197.07000732421875)
    assert node.node.BottomLevel == pytest.approx(195.0500030517578)
    assert node.node.XCoordinate == pytest.approx(-687934.6000976562)


def test_reaches_dict_access_maintains_backwards_compatibility(res1d_network, res1d_river_network):
    with pytest.warns(UserWarning):
        # Indexing reaches could return a single dotnet reach
        reach = res1d_network.result_network.reaches["100l1"]
        assert reach.Name == "100l1"
        assert reach.Length == pytest.approx(47.6827148432828)
        # Or it could include multiple reaches
        reach = res1d_river_network.result_network.reaches["river"]
        assert iter(reach), "Should be iterable where there is several subreaches"
        assert reach[0].Id == "river-12"



import warnings

from .result_location import ResultLocation
from .result_gridpoint import ResultGridPoint
from .various import make_proper_variable_name

from DHI.Mike1D.ResultDataAccess import Res1DGridPoint


class ResultReach(ResultLocation):
    """
    Class for wrapping a list of ResultData reaches
    having the same reach name.

    Parameters
    ----------
    reaches: list of IRes1DReach
        A list of MIKE 1D IRes1DReach objects having the same reach name.
    res1d : Res1D
        Res1D object the reach belongs to.

    Attributes
    ----------
    chainage_label : str
        A label, which is appended to all chainage attributes
        The value used is chainage_label = 'm_'
    data_items : list of IDataItems objects.
        A list of MIKE 1D IDataItems objects corresponding to a given reach name.
    result_gridpoints : list of lists of ResultGridPoint objects.
        A list containing lists of ResultGridPoint objects.
        Every list of ResultGridPoint objects correspond to a unique IRes1DReach.
    current_reach_result_gridpoints : list of ResultGridPoint objects.
        List of ResultGridPoint objects of currently processed IRes1DReach object.
        It is updated in set_gridpoints method.
    """

    def __init__(self, reaches, res1d):
        data_items = []
        ResultLocation.__init__(self, data_items, res1d)

        self.chainage_label = "m_"

        self.result_gridpoints = []
        self.current_reach_result_gridpoints = None

        self.reaches = []
        for reach in reaches:
            self.add_res1d_reach(reach)

    def __repr__(self) -> str:
        return f"<Reach: {self.name}>"

    def __getattr__(self, name: str):
        # TODO: Remove this in 1.0.0
        if hasattr(self.reaches[0], name):
            warnings.warn(
                f"Accessing IRes1DReach attribute {name} like this is deprecated. Use static attributes instead, or .reaches[0].{name}."
            )
            return getattr(self.reaches[0], name)
        else:
            object.__getattribute__(self, name)

    def __getitem__(self, index):
        return self.reaches[index]

    def _get_total_length(self):
        total_length = 0
        for reach in self.reaches:
            total_length += reach.Length
        return total_length

    def set_static_attributes(self):
        """Set static attributes. These show up in the html repr."""
        self.set_static_attribute("name", self.reaches[0].Name)
        self.try_set_static_attribute_length()
        self.set_static_attribute("start_chainage", self.reaches[0].LocationSpan.StartChainage)
        self.set_static_attribute("end_chainage", self.reaches[-1].LocationSpan.EndChainage)
        self.set_static_attribute("n_gridpoints", len(self.result_gridpoints))

    def try_set_static_attribute_length(self):
        try:
            self.set_static_attribute("length", self._get_total_length())
        except Exception as _:
            warnings.warn(
                "Length attribute not included. For SWMM and EPANET results this is not implemented."
            )

    def add_res1d_reach(self, reach):
        """
        Add a IRes1DReach to ResultReach.

        Parameters
        ----------
        reach: IRes1DReach
            A MIKE 1D IRes1DReach object.
        """
        self.reaches.append(reach)
        self.data_items.append(reach.DataItems)
        self.set_gridpoints(reach)
        self.set_gridpoint_data_items(reach)
        for result_gridpoint in self.current_reach_result_gridpoints:
            result_gridpoint.set_quantities()
        self.set_static_attributes()

    def set_gridpoints(self, reach):
        """
        Assign chainage attributes to a current ResultReach object
        from a data provided by IRes1DReach.

        Parameters
        ----------
        reach: IRes1DReach
            A MIKE 1D IRes1DReach object.
        """
        current_reach_result_gridpoints = []
        self.current_reach_result_gridpoints = current_reach_result_gridpoints
        self.result_gridpoints.append(current_reach_result_gridpoints)

        # For SWMM and EPANET results no grid points are defined
        # so introduce a single one.
        gridpoint_count = reach.GridPoints.Count
        if gridpoint_count == 0:
            gridpoint = Res1DGridPoint()
            self.set_gridpoint(reach, gridpoint)

        gridpoints = list(reach.GridPoints)
        for i in range(gridpoint_count):
            gridpoint = gridpoints[i]
            self.set_gridpoint(reach, gridpoint)

    def set_gridpoint(self, reach, gridpoint):
        """
        Assign chainage attribute to a current ResultReach object
        from a data provided by IRes1DReach and IRes1DGridPoint.

        Parameters
        ----------
        reach: IRes1DReach
            A MIKE 1D IRes1DReach object.
        gridpoint: IRes1DGridPoint
            A MIKE 1D IRes1DGridPoint object.
        """
        current_reach_result_gridpoints = self.current_reach_result_gridpoints

        result_gridpoint = ResultGridPoint(reach, gridpoint, reach.DataItems, self, self.res1d)
        current_reach_result_gridpoints.append(result_gridpoint)

        chainage_string = f"{gridpoint.Chainage:g}"
        result_gridpoint_attribute_string = make_proper_variable_name(
            chainage_string, self.chainage_label
        )
        setattr(self, result_gridpoint_attribute_string, result_gridpoint)

    def set_gridpoint_data_items(self, reach):
        """
        Assign data items to ResultGridPoint object belonging to current ResultReach
        from IRes1DReach data items.

        Parameters
        ----------
        reach: IRes1DReach
            A MIKE 1D IRes1DReach object.
        """
        for data_item in reach.DataItems:
            # For SWMM and EPANET results IndexList is None.
            index_list = [0] if data_item.IndexList is None else list(data_item.IndexList)
            element_count = len(index_list)
            for element_index in range(element_count):
                gridpoint_index = index_list[element_index]
                result_gridpoint = self.current_reach_result_gridpoints[gridpoint_index]
                if data_item.ItemId is None:
                    result_gridpoint.add_data_item(data_item, element_index)
                else:
                    result_gridpoint.add_structure_data_item(data_item)



import os
import sys
from glob import glob

from ..mikepath import MikePath

from .library_loader import LibraryLoader


class LibraryLoaders:
    """
    Helper class which finds all MIKE software .NET libraries
    and assigns references to those libraries in MIKE.NET module.

    Parameters
    ----------
    mikenet_module: module
        Reference to MIKE.NET module
    """

    def __init__(self, mikenet_module):
        self.mikenet_module = mikenet_module
        self.library_loader_list = []
        self.library_loader_dict = {}

        MikePath.setup_mike_installation(sys.path)
        self.create_loaders()

    def create_loaders(self):
        """
        Creates .NET library loaders and assigns relevant `load` methods
        to MIKE.NET module.
        """
        mikenet_module = self.mikenet_module

        for pattern in MikePath.library_patterns:
            file_candidates = glob(os.path.join(MikePath.mike_bin_path, pattern))
            for file_name in file_candidates:
                library_loader = LibraryLoader(file_name, mikenet_module)
                self.library_loader_list.append(library_loader)
                self.library_loader_dict[library_loader.library_name] = library_loader
                self.library_loader_dict[library_loader.library_alias] = library_loader

        setattr(mikenet_module, "load_all", self.load_all)
        setattr(mikenet_module, "load", self.load)

    def load_all(self):
        """
        Loads all libraries present in library_loader_list.
        """
        for library_loader in self.library_loader_list:
            library_loader.load()

    def load(self, libraries=[]):
        """
        Loads all libraries specified by the list `libraries`.

        Parameters
        ----------
        libraries: list of str
            List of strings specifying what .NET libraries to load.
            This can also be set as a string having the name of the
            library to load.
        """
        if isinstance(libraries, str):
            libraries = [libraries]

        for library in libraries:
            self.library_loader_dict[library].load()



import numpy as np
import pandas as pd


class ResultWriter:
    """
    Class for modifying the ResultData object TimeData
    using the value of Pandas data frame.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the modified ResultData belongs to.

    Attributes
    ----------
    result_network : ResultNetwork
        ResultNetwork object corresponding to res1d.
    """

    def __init__(self, res1d):
        self.res1d = res1d
        self.result_network = res1d.result_network

    def modify(self, data_frame):
        """
        Modifies the ResultData object TimeData based on the provided data frame.
        It will override the relevant TimeData values by the values of the data frame.

        Parameters
        ----------
        data_frame : pandas.DataFrame
            Pandas data frame object with column names based on query labels
        """
        result_quantity_map = self.result_network.result_quantity_map

        for query_label in data_frame:
            result_quantity = result_quantity_map[query_label]
            element_index = result_quantity.element_index
            data_item = result_quantity.data_item

            values = data_frame[query_label]
            time_index = data_frame.index
            self.set_values(time_index, values, data_item, element_index)

    def set_values(self, time_index, values, data_item, element_index):
        """
        Modifies the TimeData value of the data item for given element index.

        Parameters
        ----------
        time_index : pandas.DatetimeIndex
            Time stamps for the values.
        values : numpy.array
            NumPy array of values to write to TimeData.
        data_item : IDataItem
            MIKE 1D IDataItem object.
        element_index : int
            Element index into data item.
        """
        if len(values) == data_item.TimeData.NumberOfTimeSteps:
            self.set_values_all(values.values, data_item, element_index)
        else:
            self.set_values_indexed(time_index, values, data_item, element_index)

    def set_values_all(self, values, data_item, element_index):
        """
        Sets the values of all entries in TimeData
        of the data item for given element index.
        """

        # Pick the first available value.
        # TODO: Some of the query IDs can be not unique. Figure out how to handle this case.
        values_count = len(values)
        is_float = not isinstance(values[0] if values_count > 0 else 0.0, np.ndarray)

        for i in range(values_count):
            value = float(values[i] if is_float else values[i][0])
            data_item.TimeData.SetValue(i, element_index, value)

    def set_values_indexed(self, time_index, values, data_item, element_index):
        """
        Sets only the provided for time_index and values in TimeData
        of the data item for given element index.
        """
        res1d_time_index = self.res1d.time_index

        # Pick just the first column.
        # TODO: Figure out how to handle multiple column case.
        if isinstance(values, pd.DataFrame):
            values = values.iloc[:, 0]

        for time in time_index:
            value = float(values[time])
            i = res1d_time_index.get_loc(time)
            data_item.TimeData.SetValue(i, element_index, value)



import os.path

from .dotnet import from_dotnet_datetime
from .dotnet import to_dotnet_datetime
from .dotnet import to_numpy

from .result_extractor import ExtractorCreator
from .result_extractor import ExtractorOutputFileType
from .result_network import ResultNetwork
from .result_network import ResultCatchment
from .result_reader_writer import ResultReaderCreator
from .result_reader_writer import ResultReaderType
from .result_reader_writer import ResultWriter

from .query import QueryData  # noqa: F401
from .query import QueryDataCatchment  # noqa: F401
from .query import QueryDataNode  # noqa: F401
from .query import QueryDataReach  # noqa: F401
from .query import QueryDataStructure  # noqa: F401
from .query import QueryDataGlobal  # noqa: F401

from .various import mike1d_quantities  # noqa: F401
from .various import NAME_DELIMITER

from System import DateTime


class Res1D:
    """
    Class for reading data from 1D network result files into Pandas data frame.
    Currently supported formats are:
        * MIKE 1D network and catchment res1d files
        * MIKE 1D Long Term Statistics res1d files
        * MOUSE legacy PRF and CRF files
        * EPANET res and resx files generated by MIKE+
        * SWMM out files

    Parameters
    ----------
    file_path : str
        Relative or absolute path of the relevant result file.
    lazy_load : bool
        Flag specifying to load the file using lazy loading bridge of MIKE 1D.
        This typically is useful if only a few time steps need to be read for the whole network.
    header_load : bool
        Flag specifying to load just a header of the result file.
    reaches : list of str
        Filter list of reach ID strings, which will be included when loading the result file.
    nodes : list of str
        Filter list of node ID strings, which will be included when loading the result file.
    catchments : list of str
        Filter list of catchment ID strings, which will be included when loading the result file.
    col_name_delimiter : str
        String to delimit the quantity ID with location ID
        (and optionally chainage) in the data frame label.
    put_chainage_in_col_name : bool
        Flag specifying to add chainage into data frame column label.
    clear_queries_after_reading : bool
        Flag specifying to clear active queries after reading/processing them.

    Examples
    --------
    An example of reading the res1d file only for nodes with
    ID 'node1', 'node2' and reaches with ID 'reach1', 'reach2':
    ```python
    >>> nodes = ['node1', 'node2']
    >>> reaches = ['reach1', 'reach2']
    >>> res1d = Res1D('MyRes1D.res1d', nodes=nodes, reaches=reaches)
    >>> res1d.read()
    ```
    """

    def __init__(
        self,
        file_path=None,
        lazy_load=False,
        header_load=False,
        reaches=None,
        nodes=None,
        catchments=None,
        col_name_delimiter=NAME_DELIMITER,
        put_chainage_in_col_name=True,
        clear_queries_after_reading=True,
        result_reader_type=ResultReaderType.COPIER,
    ):
        self.result_reader = ResultReaderCreator.create(
            result_reader_type,
            self,
            file_path,
            lazy_load,
            header_load,
            reaches,
            nodes,
            catchments,
            col_name_delimiter,
            put_chainage_in_col_name,
        )

        self._start_time = None
        self._end_time = None

        self.result_network = ResultNetwork(self)
        self.result_writer = ResultWriter(self)
        self._queries = self.result_network.queries

        self.clear_queries_after_reading = clear_queries_after_reading

        self.catchments = self.result_network.catchments
        self.reaches = self.result_network.reaches
        self.nodes = self.result_network.nodes
        self.structures = self.result_network.structures
        self.global_data = self.result_network.global_data

    def __repr__(self):
        out = ["<mikeio1d.Res1D>"]

        if self.file_path:
            out.append(f"Start time: {str(self.start_time)}")
            out.append(f"End time: {str(self.end_time)}")
            out.append(f"# Timesteps: {str(self.data.NumberOfTimeSteps)}")
            out.append(f"# Catchments: {self.data.Catchments.get_Count()}")
            out.append(f"# Nodes: {self.data.Nodes.get_Count()}")
            out.append(f"# Reaches: {self.data.Reaches.get_Count()}")

            out.append(f"# Globals: {self.data.GlobalData.DataItems.Count}")
            for i, quantity in enumerate(self.data.Quantities):
                out.append(f"{i} - {quantity.Id} <{quantity.EumQuantity.UnitAbbreviation}>")

        return str.join("\n", out)

    # region Private methods

    def _get_actual_queries(self, queries):
        """Finds out which list of queries should be used."""
        queries = self._queries if queries is None else queries
        queries = queries if isinstance(queries, list) else [queries]
        return queries

    # endregion Private methods

    def read(self, queries=None):
        """
        Read loaded .res1d file data based on queries.
        Currently the supported query classes are
            * :class:`query.QueryDataNode`
            * :class:`query.QueryDataReach`
            * :class:`query.QueryDataCatchment`

        Parameters
        ----------
        queries: A single query or a list of queries.
            Default is None = reads all data.

        Examples
        --------
        An example of reading res1d file with queries:
        ```python
        >>> res1d = Res1D('MyRes1D.res1d')
        >>> queries = [
                QueryDataNode('WaterLevel', 'node1'),
                QueryDataReach('Discharge', 'reach1', 50.0)
            ]
        >>> res1d.read(queries)
        ```
        """

        if queries is None and len(self._queries) == 0:
            return self.read_all()

        queries = self._get_actual_queries(queries)

        df = self.result_reader.read(queries)

        return df

    def read_all(self):
        """Read all data from res1d file to dataframe."""
        return self.result_reader.read_all()

    def clear_queries(self):
        """Clear the current active list of queries."""
        self.result_network.queries.clear()
        self.result_network.queries_ids.clear()

    @property
    def time_index(self):
        """pandas.DatetimeIndex of the time index."""
        return self.result_reader.time_index

    @property
    def start_time(self):
        if self._start_time is not None:
            return self._start_time

        return from_dotnet_datetime(self.data.StartTime)

    @property
    def end_time(self):
        if self._end_time is not None:
            return self._end_time

        return from_dotnet_datetime(self.data.EndTime)

    @property
    def quantities(self):
        """Quantities in res1d file."""
        return self.result_reader.quantities

    @property
    def query(self):
        """
        .NET object ResultDataQuery to use for querying the loaded res1d data.

        More information about ResultDataQuery class see:
        https://manuals.mikepoweredbydhi.help/latest/General/Class_Library/DHI_MIKE1D/html/T_DHI_Mike1D_ResultDataAccess_ResultDataQuery.htm
        """
        return self.result_reader.query

    @property
    def searcher(self):
        """
        .NET object ResultDataSearcher to use for searching res1d data items on network.

        More information about ResultDataSearcher class see:
        https://manuals.mikepoweredbydhi.help/latest/General/Class_Library/DHI_MIKE1D/html/T_DHI_Mike1D_ResultDataAccess_ResultDataQuery.htm
        """
        return self.result_reader.searcher

    @property
    def file_path(self):
        """File path of the result file."""
        return self.result_reader.file_path

    @property
    def data(self):
        """
        .NET object ResultData with the loaded res1d data.

        More information about ResultData class see:
        https://manuals.mikepoweredbydhi.help/latest/General/Class_Library/DHI_MIKE1D/html/T_DHI_Mike1D_ResultDataAccess_ResultData.htm
        """
        return self.result_reader.data

    # region Query wrappers

    def get_catchment_values(self, catchment_id, quantity):
        return to_numpy(self.query.GetCatchmentValues(catchment_id, quantity))

    def get_node_values(self, node_id, quantity):
        return to_numpy(self.query.GetNodeValues(node_id, quantity))

    def get_reach_values(self, reach_name, chainage, quantity):
        return to_numpy(self.query.GetReachValues(reach_name, chainage, quantity))

    def get_reach_value(self, reach_name, chainage, quantity, time):
        if self.result_reader.is_lts_result_file():
            raise NotImplementedError("The method is not implemented for LTS event statistics.")

        time_dotnet = time if isinstance(time, DateTime) else to_dotnet_datetime(time)
        return self.query.GetReachValue(reach_name, chainage, quantity, time_dotnet)

    def get_reach_start_values(self, reach_name, quantity):
        return to_numpy(self.query.GetReachStartValues(reach_name, quantity))

    def get_reach_end_values(self, reach_name, quantity):
        return to_numpy(self.query.GetReachEndValues(reach_name, quantity))

    def get_reach_sum_values(self, reach_name, quantity):
        return to_numpy(self.query.GetReachSumValues(reach_name, quantity))

    # endregion Query wrapper

    def modify(self, data_frame, file_path=None):
        """
        Modifies the ResultData object TimeData based on the provided data frame.

        Parameters
        ----------
        data_frame : pandas.DataFrame
            Pandas data frame object with column names based on query labels
        file_path : str
            File path for the new res1d file. Optional.
        """
        self.result_writer.modify(data_frame)
        if file_path is not None:
            self.save(file_path)

    def save(self, file_path):
        """
        Saves the ResultData to a new res1d file.

        Parameters
        ----------
        file_path : str
            File path for the new res1d file.
        """
        self.data.Connection.FilePath.Path = file_path
        self.data.Save()

    def extract(self, file_path, queries=None, time_step_skipping_number=1, ext=None):
        """
        Extract given queries to provided file.
        File type is determined from file_path extension.
        The supported formats are:
        * csv
        * dfs0
        * txt

        Parameters
        ----------
        file_path : str
            Output file path.
        queries : list
            List of queries.
        time_step_skipping_number : int
            Number specifying the time step frequency to output.
        ext : str
            Output file type to use instead of determining it from extension.
            Can be 'csv', 'dfs0', 'txt'.
        """
        ext = os.path.splitext(file_path)[-1] if ext is None else ext

        queries = self._get_actual_queries(queries)
        data_entries = self.result_network.convert_queries_to_data_entries(queries)

        extractor = ExtractorCreator.create(
            ext, file_path, data_entries, self.data, time_step_skipping_number
        )
        extractor.export()

        if self.clear_queries_after_reading:
            self.clear_queries()

    def to_csv(self, file_path, queries=None, time_step_skipping_number=1):
        """Extract to csv file."""
        self.extract(file_path, queries, time_step_skipping_number, ExtractorOutputFileType.CSV)

    def to_dfs0(self, file_path, queries=None, time_step_skipping_number=1):
        """Extract to dfs0 file."""
        self.extract(file_path, queries, time_step_skipping_number, ExtractorOutputFileType.DFS0)

    def to_txt(self, file_path, queries=None, time_step_skipping_number=1):
        """Extract to txt file."""
        self.extract(file_path, queries, time_step_skipping_number, ExtractorOutputFileType.TXT)



from .result_locations import ResultLocations
from .result_reach import ResultReach
from .various import make_proper_variable_name
from ..dotnet import pythonnet_implementation as impl


class ResultReaches(ResultLocations):
    """
    Class for wrapping ResultData reaches.

    By itself it is also a dict, which contains
    mapping between reach name and IRes1DReach object
    or a list of IRes1DReach objects.


    Parameters
    ----------
    res1d : Res1D
        Res1D object the reaches belong to.

    Attributes
    ----------
    reach_label : str
        A label, which is appended if the reach name starts
        with a number. The value used is reach_label = 'r_'
    result_reach_map : dict
        Dictionary from reach name to a list of ResultReach objects.
        This is needed, because the reach name is not necessarily unique and
        several reaches could have the same name.
    """

    def __init__(self, res1d):
        ResultLocations.__init__(self, res1d)
        self.reach_label = "r_"
        self.result_reach_map = {}

        res1d.result_network.reaches = self
        self.set_reaches()
        self.set_quantity_collections()

    def set_reaches(self):
        """
        Set attributes to the current ResultReaches object based
        on the reach name.
        """
        for reach in self.data.Reaches:
            reach = impl(reach)
            result_reach = self.get_or_create_result_reach(reach)
            result_reach_attribute_string = make_proper_variable_name(reach.Name, self.reach_label)
            setattr(self, result_reach_attribute_string, result_reach)

    def set_quantity_collections(self):
        ResultLocations.set_quantity_collections(self)
        for reach_name in self:
            result_reach = self[reach_name]
            ResultLocations.set_quantity_collections(result_reach)

    def get_or_create_result_reach(self, reach):
        """
        Create or get already existing ResultReach object.
        There potentially could be just a single ResultReach object,
        for many IRes1DReach object, which have the same name.

        Also update self's dict entry from reach name
        to a ResultReach object.
        """
        if reach.Name in self:
            result_reach = self[reach.Name]
            result_reach.add_res1d_reach(reach)
            return result_reach

        result_reach = ResultReach([reach], self.res1d)
        self[reach.Name] = result_reach
        return result_reach



import pytest

from pandas.testing import assert_index_equal
from pandas.testing import assert_series_equal

from mikeio1d.res1d import Res1D
from mikeio1d.result_network import ResultNode
from mikeio1d.result_network import ResultCatchment
from mikeio1d.dotnet import pythonnet_implementation as impl

from .testdata import testdata


class Helpers:
    """
    Class containing helper methods for performing tests.
    """

    @staticmethod
    def assert_shared_columns_equal(df_ref, df):
        """
        Compares columns in df to the ones in df_ref.

        Note that df_ref typically has more columns than df.
        Comparison is performed only in columns of df.
        """
        assert_index_equal(df_ref.index, df.index)
        for col in df:
            # TODO: Replace with assert_series_equal(df[col], df_ref[col]) - this fails now since columns are not guaranteed unique
            diff = (df[col] - df_ref[col]).abs().sum()

            # TODO: Handle cases of different types than float
            try:
                diff = float(diff)
            except:
                continue

            assert pytest.approx(diff) == 0.0


@pytest.fixture
def helpers():
    return Helpers


@pytest.fixture()
def flow_split_file_path():
    return testdata.FlowSplit_res1d


@pytest.fixture()
def res1d_network():
    return Res1D(testdata.Network_res1d)


@pytest.fixture()
def res1d_river_network():
    return Res1D(testdata.NetworkRiver_res1d)


@pytest.fixture()
def res1d_catchments():
    return Res1D(testdata.Catchments_res1d)


@pytest.fixture
def node(res1d_network):
    dotnet_node = impl(res1d_network.data.Nodes[0])
    return ResultNode(dotnet_node, res1d_network)


@pytest.fixture
def many_nodes(res1d_network):
    nodes = res1d_network.result_network.nodes
    return [getattr(nodes, n) for n in nodes.__dict__.keys() if n.startswith(nodes.node_label)]


@pytest.fixture
def reach(res1d_network):
    return res1d_network.result_network.reaches.r_100l1


@pytest.fixture
def river_reach(res1d_river_network):
    return res1d_river_network.result_network.reaches.river


@pytest.fixture
def many_reaches(res1d_network):
    reaches = res1d_network.result_network.reaches
    return [
        getattr(reaches, r) for r in reaches.__dict__.keys() if r.startswith(reaches.reach_label)
    ]


@pytest.fixture
def catchment(res1d_catchments):
    dotnet_catchment = impl(res1d_catchments.data.Catchments[0])
    return ResultCatchment(dotnet_catchment, res1d_catchments)


@pytest.fixture
def many_catchments(res1d_catchments):
    catchments = res1d_catchments.result_network.catchments
    return [
        getattr(catchments, c)
        for c in catchments.__dict__.keys()
        if c.startswith(catchments.catchment_label)
    ]


@pytest.fixture
def structure(res1d_network):
    return res1d_network.result_network.structures["119w1"]



class DataEntry(object):
    """Class storing a Mike1D data item and a corresponding element index"""

    def __init__(self, data_item, element_index):
        self.data_item = data_item
        self.element_index = element_index



import os
import pytest

from mikeio1d import Res1D
from mikeio1d.result_network.various import make_proper_variable_name


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # Original file name was Exam6Base.res1d
    return os.path.join(test_folder_path, "testdata", "Network_chinese.res1d")


def test_make_proper_variable_name():
    mpvn = make_proper_variable_name  # alias
    assert mpvn("a") == "a"
    assert mpvn("1") == "_1"
    assert mpvn("1a") == "_1a"
    assert mpvn("a1") == "a1"
    assert mpvn("a1a") == "a1a"
    assert mpvn("myname??++") == "myname"
    assert mpvn("myname??++something") == "myname_something"
    assert mpvn("你好") == "你好"
    assert mpvn("123你好") == "_123你好"


def test_network_attributes_allow_chinese_characters(test_file_path):
    """Test that network attributes allow chinese characters."""
    res = Res1D(test_file_path)
    assert res.nodes.风景
    assert res.nodes.电脑
    assert res.nodes.餐厅
    assert res.nodes.爱情
    assert res.nodes.音乐
    assert res.nodes.星期
    assert res.nodes.医生
    assert res.nodes.美丽
    assert res.nodes.学生
    assert res.nodes.蛋糕



import os
import pytest
import numpy as np
import pandas as pd
import datetime

from mikeio1d.custom_exceptions import NoDataForQuery, InvalidQuantity
from mikeio1d.res1d import Res1D
from mikeio1d.query import QueryDataReach, QueryDataNode, QueryDataGlobal
from mikeio1d.dotnet import to_numpy


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # Original file name was Exam6Base.res1d
    return os.path.join(test_folder_path, "testdata", "LTSMonthlyStatistics.res1d")


@pytest.fixture(params=[False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_file_does_not_exist():
    with pytest.raises(FileExistsError):
        assert Res1D("tests/testdata/not_a_file.res1d")


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 73
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 21


def test_repr(test_file):
    res1d = test_file
    res1d_repr = res1d.__repr__()
    res1d_repr_ref = (
        "<mikeio1d.Res1D>\n"
        + "Start time: 1957-01-01 00:00:00\n"
        + "End time: 1963-01-01 00:00:00\n"
        + "# Timesteps: 73\n"
        + "# Catchments: 0\n"
        + "# Nodes: 16\n"
        + "# Reaches: 17\n"
        + "# Globals: 9\n"
        + "0 - DischargeIntegratedMonthly <m^3>\n"
        + "1 - DischargeIntegratedMonthlyCount <()>\n"
        + "2 - DischargeIntegratedMonthlyDuration <h>\n"
        + "3 - Component_1TransportIntegratedMonthly <kg>\n"
        + "4 - Component_1TransportIntegratedMonthlyCount <()>\n"
        + "5 - Component_1TransportIntegratedMonthlyDuration <h>\n"
        + "6 - Component_2TransportIntegratedMonthly <kg>\n"
        + "7 - Component_2TransportIntegratedMonthlyCount <()>\n"
        + "8 - Component_2TransportIntegratedMonthlyDuration <h>\n"
        + "9 - SurchargeIntegratedMonthly <m^3>\n"
        + "10 - SurchargeIntegratedMonthlyCount <()>\n"
        + "11 - SurchargeIntegratedMonthlyDuration <h>\n"
        + "12 - DischargeIntegratedMonthlyOutlets <m^3>\n"
        + "13 - DischargeIntegratedMonthlyWeirs <m^3>\n"
        + "14 - DischargeIntegratedMonthlyTotalOutflow <m^3>\n"
        + "15 - Component_1TransportIntegratedMonthlyTotalEmission <kg>\n"
        + "16 - Component_2TransportIntegratedMonthlyTotalEmission <kg>\n"
        + "17 - Component_1TransportIntegratedMonthlyOutlets <kg>\n"
        + "18 - Component_2TransportIntegratedMonthlyOutlets <kg>\n"
        + "19 - Component_1TransportIntegratedMonthlyWeirs <kg>\n"
        + "20 - Component_2TransportIntegratedMonthlyWeirs <kg>"
    )
    assert res1d_repr == res1d_repr_ref


def test_data_item_dicts(test_file):
    res1d = test_file
    assert len(res1d.catchments) == 0
    assert len(res1d.nodes) == 16
    assert len(res1d.reaches) == 17
    assert len(res1d.global_data) == 9


@pytest.mark.parametrize(
    "query,expected",
    [
        (QueryDataReach("DischargeIntegratedMonthly", "B4.1320l1", 0), True),
        (QueryDataReach("DischargeIntegratedMonthly", "B4.1320l1", 42424242), False),
        (QueryDataReach("DischargeIntegratedMonthly", "wrong_reach_name", 0), False),
    ],
)
def test_valid_reach_data_queries(test_file, query, expected):
    res1d = test_file

    with pytest.raises(InvalidQuantity):
        invalid_query = QueryDataReach("InvalidQuantity", "104l1", 34.4131)
        assert res1d.read(invalid_query)

    if expected:
        res1d.read(query)
    else:
        with pytest.raises(NoDataForQuery):
            assert res1d.read(query)


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataReach("DischargeIntegratedMonthly", "B4.1320l1", 135.001), 1215.915),
        (QueryDataReach("DischargeIntegratedMonthly", "B4.1491l1", 144), 563.973),
        (QueryDataReach("DischargeIntegratedMonthlyCount", "B4.1320l1", 270.002), 3),
        (QueryDataReach("DischargeIntegratedMonthlyCount", "B4.1491l1", 239.999), 3),
    ],
)
def test_read_reach_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


@pytest.mark.parametrize(
    "quantity,reach_id,chainage,expected_max",
    [
        ("DischargeIntegratedMonthly", "B4.1320l1", 135.001, 1215.915),
        ("DischargeIntegratedMonthly", "B4.1491l1", 144, 563.973),
        ("DischargeIntegratedMonthlyCount", "B4.1320l1", 270.002, 3),
        ("DischargeIntegratedMonthlyCount", "B4.1491l1", 239.999, 3),
    ],
)
def test_read_reach(test_file, quantity, reach_id, chainage, expected_max):
    data = test_file.query.GetReachValues(reach_id, chainage, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


@pytest.mark.parametrize(
    "quantity,node_id,expected_max",
    [
        ("SurchargeIntegratedMonthly", "B4.1200", 97.434),
        ("SurchargeIntegratedMonthlyCount", "B4.1200", 1),
    ],
)
def test_read_node(test_file, quantity, node_id, expected_max):
    data = test_file.query.GetNodeValues(node_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


def test_time_index(test_file):
    assert len(test_file.time_index) == 73


def test_start_time(test_file):
    time = test_file.start_time
    date = datetime.datetime(time.year, time.month, 1, 0, 0)
    assert date == test_file.time_index.min()


def test_get_node_values(test_file):
    values = test_file.get_node_values("B4.1200", "SurchargeIntegratedMonthly")
    assert len(values) == 73


def test_get_reach_values(test_file):
    values = test_file.get_reach_values("B4.1491l1", 144, "DischargeIntegratedMonthlyCount")
    time_series = pd.Series(values, index=test_file.time_index)
    assert len(values) == 73
    assert len(time_series.index) == 73
    # Just try to call the methods
    test_file.get_reach_end_values("B4.1491l1", "DischargeIntegratedMonthlyCount")
    test_file.get_reach_start_values("B4.1491l1", "DischargeIntegratedMonthlyCount")
    test_file.get_reach_sum_values("B4.1491l1", "DischargeIntegratedMonthlyCount")


def test_get_reach_value(test_file):
    value = test_file.get_reach_value(
        "B4.1491l1", 144, "DischargeIntegratedMonthlyCount", test_file.start_time
    )
    assert value == 0


def test_res1d_filter(test_file_path, helpers):
    nodes = ["B4.1320", "A0.0327"]
    reaches = ["B4.1491l1"]
    res1d = Res1D(test_file_path, nodes=nodes, reaches=reaches)

    df_b4_14 = res1d.read(QueryDataReach("DischargeIntegratedMonthly", "B4.1491l1", 144))
    df_b4_13 = res1d.read(QueryDataNode("SurchargeIntegratedMonthly", "B4.1320"))
    df_a0 = res1d.read(QueryDataNode("DischargeIntegratedMonthly", "A0.0327"))

    res1d_full = Res1D(test_file_path)
    df_full = res1d_full.read()

    helpers.assert_shared_columns_equal(df_full, df_b4_14)
    helpers.assert_shared_columns_equal(df_full, df_b4_13)
    helpers.assert_shared_columns_equal(df_full, df_a0)

    # Release the .NET object
    res1d = None


def test_res1d_filter_readall(test_file_path, helpers):
    # Make sure read all can be used with filters
    nodes = ["B4.1320", "A0.0327"]
    reaches = ["B4.1491l1"]
    res1d = Res1D(test_file_path, nodes=nodes, reaches=reaches)
    df = res1d.read()

    res1d_full = Res1D(test_file_path)
    df_full = res1d_full.read()

    helpers.assert_shared_columns_equal(df_full, df)

    # Release the .NET object
    res1d = None


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataGlobal("DischargeIntegratedMonthlyOutlets"), 5562.719),
        (QueryDataGlobal("DischargeIntegratedMonthlyWeirs"), 437.729),
        (QueryDataGlobal("DischargeIntegratedMonthlyTotalOutflow"), 5971.352),
    ],
)
def test_read_global_items_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


def test_global_data_attributes(test_file):
    res1d = test_file
    global_data = res1d.global_data

    global_data.DischargeIntegratedMonthlyOutlets.add()
    global_data.DischargeIntegratedMonthlyWeirs.add()
    global_data.DischargeIntegratedMonthlyTotalOutflow.add()

    df = res1d.read()

    actual_max = round(df["DischargeIntegratedMonthlyOutlets"].max(), 3)
    assert pytest.approx(actual_max) == 5562.719

    actual_max = round(df["DischargeIntegratedMonthlyWeirs"].max(), 3)
    assert pytest.approx(actual_max) == 437.729

    actual_max = round(df["DischargeIntegratedMonthlyTotalOutflow"].max(), 3)
    assert pytest.approx(actual_max) == 5971.352



from .result_locations import ResultLocations
from .result_structure import ResultStructure
from .various import make_proper_variable_name


class ResultStructures(ResultLocations):
    """
    Class for wrapping ResultData structure data items.

    By itself it is also a dict, which contains
    mapping between structure ID and a ResultStructure object.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the structure data items belong to.

    Attributes
    ----------
    structure_label : str
        A label, which is appended if the structure ID starts
        with a number. The value used is structure_label = 's_'
    result_structure_map : dict
        Dictionary from structure ID to a ResultStructure object.
    """

    def __init__(self, res1d):
        ResultLocations.__init__(self, res1d)
        self.structure_label = "s_"
        self.result_structure_map = {}

        res1d.result_network.structures = self
        self.set_structures()
        self.set_quantity_collections()

    def set_structures(self):
        """
        Set attributes to the current ResultReaches object based
        on the reach name.
        """
        for reach in self.data.Reaches:
            for data_item in reach.DataItems:
                if not self.is_structure(reach, data_item):
                    continue

                result_structure = self.get_or_create_result_structure(reach, data_item)
                structure_id = result_structure.id
                result_structure_attribute_string = make_proper_variable_name(
                    structure_id, self.structure_label
                )
                setattr(self, result_structure_attribute_string, result_structure)

    def is_structure(self, reach, data_item):
        # Data items on reaches with defined ItemId correspond to structure data items.
        if data_item.ItemId is not None:
            return True

        # Data item with no index list is not a structure data item.
        if data_item.IndexList is None:
            return False

        is_data_item_for_single_grid_point = len(list(data_item.IndexList)) == 1
        has_three_grid_points = len(list(reach.GridPoints)) == 3

        if reach.IsStructureReach and has_three_grid_points and is_data_item_for_single_grid_point:
            return True

        return False

    def get_or_create_result_structure(self, reach, data_item):
        """
        Create or get already existing ResultStructure object.

        Also update a result_structure_map dict entry from structure ID
        to a ResultStructure object.
        """
        structure_id = ResultStructure.get_structure_id(reach, data_item)

        result_structure_map = self.result_structure_map
        if structure_id in result_structure_map:
            result_structure = result_structure_map[structure_id]
            result_structure.add_res1d_structure_data_item(data_item)
            return result_structure

        result_structure = ResultStructure(structure_id, reach, [data_item], self.res1d)
        self.set_res1d_object_to_dict(structure_id, result_structure)
        result_structure_map[structure_id] = result_structure
        return result_structure



import numpy as np

from ..custom_exceptions import NoDataForQuery
from ..custom_exceptions import InvalidQuantity
from ..various import NAME_DELIMITER


class QueryData:
    """
    Base query class that declares what data to extract from a .res1d file.

    Parameters
    ----------
    quantity: str
        e.g. 'WaterLevel', 'Discharge', etc. Call res1d.quantities to get all quantities.
    name: str
        Name or ID of location under interest.
    validate: bool
        Flag specifying to validate the query.
    """

    delete_value = -1e-30

    def __init__(self, quantity, name=None, validate=True):
        self._name = name
        self._quantity = quantity

        if validate:
            self._validate()

    def _validate(self):
        if not isinstance(self.quantity, str):
            raise TypeError("Quantity must be a string.")

        if self.name is not None and not isinstance(self.name, str):
            raise TypeError("Argument 'name' must be either None or a string.")

    def add_to_data_entries(self, res1d, data_entries, column_names):
        self._check_invalid_quantity(res1d)

        self._update_query(res1d)

        query_label = str(self)
        result_quantity = res1d.result_network.result_quantity_map.get(query_label, None)

        self._check_invalid_values(result_quantity)

        column_names.append(query_label)

        data_entry = result_quantity.get_data_entry_net()
        data_entries.Add(data_entry)

    @staticmethod
    def from_dotnet_to_python(array):
        """Convert .NET array to numpy."""
        return np.fromiter(array, np.float64)

    @property
    def quantity(self):
        return self._quantity

    @property
    def name(self):
        return self._name

    def _update_query(self, res1d):
        pass

    def _check_invalid_quantity(self, res1d):
        if self._quantity not in res1d.quantities:
            raise InvalidQuantity(
                f"Undefined quantity {self._quantity}. "
                f"Allowed quantities are: {', '.join(res1d.quantities)}."
            )

    def _check_invalid_values(self, values):
        if values is None:
            raise NoDataForQuery(str(self))

    def __repr__(self):
        return NAME_DELIMITER.join([self._quantity, self._name])



from collections import defaultdict, namedtuple
import functools

import os.path
import pandas as pd

from DHI.Mike1D.CrossSectionModule import CrossSectionDataFactory
from DHI.Mike1D.Generic import Connection, Diagnostics, Location


class BaseXns11Error(Exception):
    """Base class for Xns11 errors."""


class DataNotFoundInFile(BaseXns11Error):
    """Data not found in file."""


class FileNotOpenedError(BaseXns11Error):
    """Accessing data from a file that is not opened."""


def read(file_path, queries):
    """Read the requested data from the xns11 file and
    return a Pandas DataFrame.

    Parameters
    ----------
    file_path: str
        full path and file name to the xns11 file.
    queries: a single query or a list of queries
        `QueryData` objects that define the requested data.
    Returns
    -------
    pd.DataFrame
    """
    queries = queries if isinstance(queries, list) else [queries]
    with open(file_path) as xns11:
        return xns11.read(queries)


def open(file_path):
    """Open a xns11 file as a Xns11 object that has convenient methods
    to extract specific data from the file. It is recommended to use it
    as a context manager.

    Parameters
    ----------
    file_path: str
        full path and file name to the xns11 file.

    Returns
    -------
    Xns11

    Examples
    --------
    >>> with open("file.xns11") as x11:
    >>>     print(x11.topoid_names)
    ['topoid1', 'topoid2']
    """
    return Xns11(file_path)


def _not_closed(prop):
    @functools.wraps(prop)
    def wrapper(self, *args, **kwargs):
        if self._closed:
            raise FileNotOpenedError(
                "The xns11 file should be opened first to access to its "
                f"{prop.__name__} property."
            )
        return prop(self, *args, **kwargs)

    return wrapper


class Xns11:
    def __init__(self, file_path=None):
        self.file_path = file_path
        self.file = None
        self._closed = True
        self._reach_names = None
        self.__reaches = None
        self._topoid_names = None
        self.__topoids = None
        # Load the file on initialization
        self._load_file()

    def _load_file(self):
        """Load the file."""
        if not os.path.exists(self.file_path):
            raise FileExistsError(f"File {self.file_path} does not exist.")
        self.file = CrossSectionDataFactory().Open(
            Connection.Create(self.file_path), Diagnostics("Error loading file.")
        )
        self._closed = False

    def close(self):
        """Close the file handle."""
        self.file.Finalize()
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, *excinfo):
        self.close()

    @property
    def _topoids(self):
        if self.__topoids:
            return self.__topoids
        return list(self.file.GetReachTopoIdEnumerable())

    @property
    @_not_closed
    def topoid_names(self):
        """A list of the topo-id names"""
        if self._topoid_names:
            return self._topoid_names
        return [topoid.TopoId for topoid in self._topoids]

    @property
    def _reaches(self):
        if self.__reaches:
            return self.__reaches
        return list(self.file.GetReachTopoIdEnumerable())

    @property
    @_not_closed
    def reach_names(self):
        """A list of the reach names"""
        if self._reach_names:
            return self._reachs_names
        return [reach.ReachId for reach in self._topoids]

    @staticmethod
    def _topoid_in_reach(self, reach):
        """A list of the topo-ID contained in a reach."""
        return [
            r.TopoId
            for r in list(self.file.GetReachTopoIdEnumerable())
            if reach.ReachId == r.ReachId
        ]

    @staticmethod
    def _chainages(reach):
        """A list of chainages of a reach topo-ID combination."""
        return [r.Key for r in list(reach.GetChainageSortedCrossSections())]

    def _get_values(self, points):
        df = pd.DataFrame()
        p = zip(points["chainage"], points["reach"], points["topoid"])
        for chainage, reach, topoid in p:
            location = Location()
            location.ID = reach.value
            location.Chainage = chainage.value
            geometry = self.file.FindClosestCrossSection(
                location, topoid.value
            ).BaseCrossSection.Points
            x, z = [], []
            for i in range(geometry.Count):
                x.append(geometry.LstPoints[i].X)
                z.append(geometry.LstPoints[i].Z)
            x_name = f"x {topoid.value} {reach.value} {chainage.value}"
            z_name = f"z {topoid.value} {reach.value} {chainage.value}"
            geometry_x = pd.Series(x, name=x_name)
            geometry_z = pd.Series(z, name=z_name)
            df = pd.concat([df, geometry_x, geometry_z], axis=1)
        return df

    def _get_data(self, points):
        df = self._get_values(points)
        return df

    def _validate_queries(self, queries, chainage_tolerance=0.1):
        """Check whether the queries point to existing data in the file."""
        for q in queries:
            if q.topoid_name not in self.topoid_names:
                raise DataNotFoundInFile(f"Topo-id '{q.topoid_name}' was not found.")
            if q.reach_name is not None:
                if q.reach_name not in self.reach_names:
                    raise DataNotFoundInFile(f"Reach '{q.reach_name}' was not found.")
                else:
                    found_topo_id_in_reach = False
                    for reach in self._reaches:
                        if found_topo_id_in_reach:
                            break
                        # Look for the targeted reach
                        if q.reach_name != reach.ReachId:
                            continue
                        # Raise an error if the combination reach and topo-id does not exist
                        topoid_names_in_reach = self._topoid_in_reach(self, reach)
                        if q.topoid_name not in topoid_names_in_reach:
                            raise DataNotFoundInFile(
                                f"Topo-ID '{q.topoid_name}' was not found in reach '{q.reach_name}'."
                            )
                        found_topo_id_in_reach = True
            if q.chainage is not None:
                found_chainage = False
                for reach in self._reaches:
                    if found_chainage:
                        break
                    # Look for the targeted combination reach and topo-id
                    if q.reach_name != reach.ReachId and q.topoid_name != reach.TopoId:
                        continue
                    for chainage in self._chainages(reach):
                        # Look for the targeted chainage
                        chainage_diff = chainage - q.chainage
                        if abs(chainage_diff) < chainage_tolerance:
                            found_chainage = True
                            break
                if not found_chainage:
                    raise DataNotFoundInFile(
                        f"Chainage {q.chainage} was not found in reach '{q.reach_name}' for Topo-ID '{q.topoid_name}'."
                    )

    def _build_queries(self, queries):
        """ "
        A query can be in an undefined state if reach_name and/or chainage
        isn't set. This function takes care of building lists of queries
        for these cases. Chainages are rounded to three decimal places.

        >>> self._build_queries([QueryData("topoid1", "reach1")])
        [
            QueryData("topoid1", "reach1", 56.68),
            QueryData("topoid1", "reach1", 121.98)
        ]
        """
        built_queries = []
        for q in queries:
            # e.g. QueryData("topoid1", "reach1", 58.68)
            if q.reach_name and q.chainage:
                built_queries.append(q)
                continue
            # e.g QueryData("topoid1", "reach1") or QueryData("topoid1")
            q_topoid_name = q.topoid_name
            q_reach_name = q.reach_name
            for reach, reach_name in zip(self._reaches, self.reach_names):
                if q_reach_name is not None:  # When reach_name is set
                    if reach_name != q_reach_name:
                        continue
                topoid_in_reach = self._topoid_in_reach(self, reach)
                if q.topoid_name not in topoid_in_reach:
                    continue
                for curr_chain in self._chainages(reach):
                    if reach.TopoId == q.topoid_name:
                        chainage = curr_chain
                    else:
                        continue

                    q = QueryData(q_topoid_name, reach_name, round(chainage, 3))
                    built_queries.append(q)
        return built_queries

    def _find_points(self, queries, chainage_tolerance=0.1):
        """
        From a list of queries returns a dictionary with the required
        information for each requested point to extract its geometry
        later on.
        """

        PointInfo = namedtuple("PointInfo", ["index", "value"])

        found_points = defaultdict(list)
        # Find the points given its topo-id, reach, and chainage
        for q in queries:
            for reach_idx, curr_reach in enumerate(self._reaches):
                # Look for the targed reach and topo-id
                if q.reach_name != curr_reach.ReachId or q.topoid_name != curr_reach.TopoId:
                    continue
                reach_info = PointInfo(reach_idx, q.reach_name)
                topo_pair = self._topoid_in_reach(self, curr_reach)
                for topoid_reach_idx, topoid_reach in enumerate(topo_pair):
                    if q.topoid_name != topoid_reach:
                        continue
                    topoid_info = PointInfo(topoid_reach_idx, topoid_reach)
                    for idx, curr_chain in enumerate(self._chainages(curr_reach)):
                        # Look for the targed chainage
                        chainage_diff = curr_chain - q.chainage
                        is_chainage = abs(chainage_diff) < chainage_tolerance
                        if not is_chainage:
                            continue
                        # idx is the index in the topoid
                        chainage_info = PointInfo(idx, q.chainage)
                        found_points["chainage"].append(chainage_info)
                        found_points["topoid"].append(topoid_info)
                        found_points["reach"].append(reach_info)
                        break  # Break at the first chainage found.

        return dict(found_points)

    @_not_closed
    def read(self, queries):
        """Read the requested data from the xns11 file and
        return a Pandas DataFrame.

        Parameters
        ----------
        queries: list
            `QueryData` objects that define the requested data.
        Returns
        -------
        pd.DataFrame
        """
        self._validate_queries(queries)
        built_queries = self._build_queries(queries)
        found_points = self._find_points(built_queries)
        df = self._get_data(found_points)
        return df


class QueryData:
    """A query object that declares what data should be
    extracted from a .xns11 file.

    Parameters
    ----------
    topoid_name: str
        Topo ID, must be passed
    reach_name: str, optional
        Reach name, consider all the reaches if None
    chainage: float, optional
        chainage, considers all the chainages if None

    Examples
    --------
    `QueryData('topoid1', 'reach1', 10)` is a valid query.
    `QueryData('topoid1', 'reach1')` requests all the points
    for `topoid1` of `reach1`.
    `QueryData('topoid1')` requests all the points for `topid1`
    of the file.
    """

    def __init__(self, topoid_name, reach_name=None, chainage=None):
        self._topoid_name = topoid_name
        self._reach_name = reach_name
        self._chainage = chainage
        self._validate()

    def _validate(self):
        tp = self.topoid_name
        rn = self.reach_name
        c = self.chainage
        if not isinstance(tp, str):
            raise TypeError("topoid_name must be a string.")
        if rn is not None and not isinstance(rn, str):
            raise TypeError("reach_name must be either None or a string.")
        if c is not None and not isinstance(c, (int, float)):
            raise TypeError("chainage must be either None or a number.")
        if rn is None and c is not None:
            raise ValueError("chainage cannot be set if reach_name is None.")

    @property
    def topoid_name(self):
        return self._topoid_name

    @property
    def reach_name(self):
        return self._reach_name

    @property
    def chainage(self):
        return self._chainage

    def __repr__(self):
        return (
            f"QueryData(topoid_name='{self.topoid_name}', "
            f"reach_name='{self.reach_name}', "
            f"chainage={self.chainage})"
        )



import os.path
import pandas as pd
import datetime

from ..dotnet import from_dotnet_datetime
from ..various import NAME_DELIMITER

from DHI.Mike1D.ResultDataAccess import ResultData
from DHI.Mike1D.ResultDataAccess import ResultDataQuery
from DHI.Mike1D.ResultDataAccess import ResultDataSearch
from DHI.Mike1D.ResultDataAccess import Filter
from DHI.Mike1D.ResultDataAccess import DataItemFilterName
from DHI.Mike1D.ResultDataAccess import ResultTypes

from DHI.Mike1D.Generic import Connection
from DHI.Mike1D.Generic import Diagnostics


class ResultReader:
    """
    Class for reading the ResultData object TimeData
    into Pandas data frame.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the modified ResultData belongs to.
    file_path : str
        Relative or absolute path of the relevant result file.
    lazy_load : bool
        Flag specifying to load the file using lazy loading bridge of MIKE 1D.
        This typically is useful if only a few time steps need to be read for the whole network.
    header_load : bool
        Flag specifying to load just a header of the result file.
    reaches : list of str
        Filter list of reach ID strings, which will be included when loading the result file.
    nodes : list of str
        Filter list of node ID strings, which will be included when loading the result file.
    catchments : list of str
        Filter list of catchment ID strings, which will be included when loading the result file.
    col_name_delimiter : str
        String to delimit the quantity ID with location ID
        (and optionally chainage) in the data frame label.
    put_chainage_in_col_name : bool
        Flag specifying to add chainage into data frame column label.
    """

    def __init__(
        self,
        res1d,
        file_path=None,
        lazy_load=False,
        header_load=False,
        reaches=None,
        nodes=None,
        catchments=None,
        col_name_delimiter=NAME_DELIMITER,
        put_chainage_in_col_name=True,
    ):
        self.res1d = res1d

        self.file_path = file_path
        self.file_extension = os.path.splitext(file_path)[-1]

        self.lazy_load = lazy_load

        self._reaches = reaches if reaches else []
        self._nodes = nodes if nodes else []
        self._catchments = catchments if catchments else []

        self.use_filter = reaches is not None or nodes is not None or catchments is not None

        self._load_header()
        if not header_load:
            self._load_file()

        self._time_index = None

        self.col_name_delimiter = col_name_delimiter
        self.put_chainage_in_col_name = put_chainage_in_col_name

        self.quantities = [quantity.Id for quantity in self.data.Quantities]

    # region File loading

    def _load_header(self):
        if not os.path.exists(self.file_path):
            raise FileExistsError(f"File {self.file_path} does not exist.")

        self.data = ResultData()
        self.data.Connection = Connection.Create(self.file_path)
        self.diagnostics = Diagnostics("Loading header")

        if self.lazy_load:
            self.data.Connection.BridgeName = "res1dlazy"

        if self.use_filter:
            self.data.LoadHeader(True, self.diagnostics)
        else:
            self.data.LoadHeader(self.diagnostics)

    def _load_file(self):
        if self.use_filter:
            self._setup_filter()

            for reach in self._reaches:
                self._add_reach(reach)
            for node in self._nodes:
                self._add_node(node)
            for catchment in self._catchments:
                self._add_catchment(catchment)

        if self.file_extension.lower() in [".resx", ".crf", ".prf", ".xrf"]:
            self.data.Load(self.diagnostics)
        else:
            self.data.LoadData(self.diagnostics)

        self.query = ResultDataQuery(self.data)
        self.searcher = ResultDataSearch(self.data)

    def _setup_filter(self):
        """
        Setup the filter for result data object.
        """
        if not self.use_filter:
            return

        self.data_filter = Filter()
        self.data_subfilter = DataItemFilterName(self.data)
        self.data_filter.AddDataItemFilter(self.data_subfilter)

        self.data.Parameters.Filter = self.data_filter

    def _add_reach(self, reach_id):
        self.data_subfilter.Reaches.Add(reach_id)

    def _add_node(self, node_id):
        self.data_subfilter.Nodes.Add(node_id)

    def _add_catchment(self, catchment_id):
        self.data_subfilter.Catchments.Add(catchment_id)

    # endregion File loading

    def read(self, queries=None):
        return None

    def read_all(self):
        return None

    def is_data_set_included(self, data_set):
        """Skip filtered data sets"""
        name = self.get_data_set_name(data_set)
        if self.use_filter and name not in self._catchments + self._reaches + self._nodes:
            return False
        return True

    @property
    def time_index(self):
        """pandas.DatetimeIndex of the time index."""
        if self._time_index is not None:
            return self._time_index

        if self.is_lts_result_file():
            return self.lts_event_index

        time_stamps = [from_dotnet_datetime(t) for t in self.data.TimesList]
        self._time_index = pd.DatetimeIndex(time_stamps)
        return self._time_index

    def get_data_set_name(self, data_set, item_id=None):
        name = None

        if hasattr(data_set, "Name"):
            name = data_set.Name
        elif hasattr(data_set, "Id"):
            name = data_set.Id
        elif data_set.Quantity is not None:
            name = data_set.Quantity.Id

        name = "" if name is None else name

        # Add item id if present before the name.
        # Needed for unique identification of structures.
        name = self.col_name_delimiter.join([item_id, name]) if item_id is not None else name

        return name

    def get_column_name(self, data_set, data_item, i):
        quantity_id = data_item.Quantity.Id
        item_id = data_item.ItemId
        name = self.get_data_set_name(data_set, item_id)

        chainage = None
        if data_item.IndexList is not None:
            chainages = data_set.GetChainages(data_item)
            chainage = chainages[i]

        if name == "":
            return quantity_id

        if chainage is None:
            return self.col_name_delimiter.join([quantity_id, name])

        postfix = f"{chainage:g}" if self.put_chainage_in_col_name else str(i)
        return self.col_name_delimiter.join([quantity_id, name, postfix])

    # region Methods for LTS result files

    def update_time_quantities(self, df):
        if not self.is_lts_result_file():
            return

        simulation_start = from_dotnet_datetime(self.data.StartTime)
        for label in df:
            time_suffix = f"Time{self.col_name_delimiter}"
            if time_suffix in label:
                seconds_after_simulation_start_array = df[label].to_numpy()
                times = [
                    simulation_start + datetime.timedelta(seconds=float(sec))
                    for sec in seconds_after_simulation_start_array
                ]
                df[label] = times

    def is_lts_result_file(self):
        # For pythonnet version > 3.0 it is possible to call
        # return self._data.ResultType.Equals(ResultTypes.LTSEvents)
        return int(self.data.ResultType) == int(ResultTypes.LTSEvents)

    @property
    def lts_event_index(self):
        """pandas.DatetimeIndex of the LTS event index."""
        if self._time_index is not None:
            return self._time_index

        number_of_event_entries = len(self.data.TimesList)
        event_index = [i for i in range(number_of_event_entries)]

        self._time_index = pd.Index(event_index)

        return self._time_index

    # endregion Methods for LTS result files



from .extractor_creator import ExtractorCreator
from .extractor_creator import ExtractorOutputFileType

from .extractor import Extractor
from .extractor_all import ExtractorAll
from .extractor_csv import ExtractorCsv
from .extractor_dfs0 import ExtractorDfs0
from .extractor_txt import ExtractorTxt



from .result_quantity import ResultQuantity


class ResultQuantityCollection(ResultQuantity):
    """
    Class for dealing with a collection ResultQuantity objects.

    ResultQuantityCollection objects are the attributes assigned to a network
    type like nodes, catchments, etc. For example, res1d.nodes.WaterLevel
    They have the ability to add queries.

    Parameters
    ----------
    result_quantities: list of ResultQuantity objects
        A list of ResultQuantity objects having the same quantity id.
    res1d : Res1D
        Res1D object the quantity belongs to.
    """

    def __init__(self, result_quantities, res1d):
        self.result_quantities = result_quantities
        self.res1d = res1d

    def add(self):
        """
        Add queries to ResultNetwork.queries from a list of result quantities.
        """
        for result_quantity in self.result_quantities:
            result_quantity.add()

    def plot(self):
        """Plot the time series data."""
        if len(self.result_quantities) <= 0:
            return

        self.data_item = self.result_quantities[0].data_item
        ResultQuantity.plot(self)

    def get_query(self):
        """Get queries corresponding to ResultQuantityCollection."""
        queries = []
        for result_quantity in self.result_quantities:
            query = result_quantity.get_query()
            queries.append(query)
        return queries



from ..dotnet import pythonnet_implementation as impl
from .result_quantity_collection import ResultQuantityCollection
from .various import make_proper_variable_name


class ResultLocations(dict):
    """
    A base class for a network locations (nodes, reaches)
    or a catchments wrapper class.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the result location belongs to.

    Attributes
    ----------
    data : ResultData
        MIKE 1D ResultData object.
    data_items : IDataItems.
        MIKE 1D IDataItems object.
    quantity_label : str
        A label, which is appended if the quantity id starts
        with a number. The value used is quantity_label = 'q_'
    result_quantity_map : dict
        Dictionary from quantity id to a list of ResultQuantity objects.
    """

    def __init__(self, res1d):
        self.res1d = res1d
        self.quantity_label = "q_"
        self.data = res1d.data
        self.data_items = res1d.data.DataItems
        self.result_quantity_map = {}

    def set_quantity_collections(self):
        """Sets all quantity collection attributes."""
        for quantity_id in self.result_quantity_map:
            result_quantities = self.result_quantity_map[quantity_id]
            result_quantity_collection = ResultQuantityCollection(result_quantities, self.res1d)
            result_quantity_attribute_string = make_proper_variable_name(
                quantity_id, self.quantity_label
            )
            setattr(self, result_quantity_attribute_string, result_quantity_collection)

    def set_res1d_object_to_dict(self, dict_key, obj):
        """
        Create a dict entry from a key name to an object
        or a list of objects.
        """
        obj = impl(obj)
        if dict_key in self:
            value = self[dict_key]
            if not isinstance(value, list):
                self[dict_key] = [value]

            self[dict_key].append(obj)
        else:
            self[dict_key] = obj



from .query_data import QueryData


class QueryDataGlobal(QueryData):
    """A query object that declares what global data to extract from a .res1d file.

    Parameters
    ----------
    quantity: str
        e.g. 'TimeStep'. Call res1d.quantities to get all quantities.
    validate: bool
        Flag specifying to validate the query.

    Examples
    --------
    `QueryDataGlobal('TimeStep')` is a valid query.
    """

    def __init__(self, quantity, validate=True):
        super().__init__(quantity, validate=validate)

    def get_values(self, res1d):
        self._check_invalid_quantity(res1d)

        data_item = res1d.global_data[self._quantity]
        values = data_item.CreateTimeSeriesData(0)

        self._check_invalid_values(values)

        return self.from_dotnet_to_python(values)

    def __repr__(self):
        return self._quantity



from ..dotnet import pythonnet_implementation as impl
from .result_locations import ResultLocations
from .result_node import ResultNode
from .various import make_proper_variable_name


class ResultNodes(ResultLocations):
    """
    Class for wrapping ResultData nodes.

    By itself it is also a dict, which contains
    mapping between node ID and IRes1DNode object.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the nodes belong to.

    Attributes
    ----------
    node_label : str
        A label, which is appended if the node name starts
        with a number. The value used is node_label = 'n_'
    """

    def __init__(self, res1d):
        ResultLocations.__init__(self, res1d)
        self.node_label = "n_"

        res1d.result_network.nodes = self
        self.set_nodes()
        self.set_quantity_collections()

    def set_nodes(self):
        """
        Set attributes to the current ResultNodes object based
        on the node ID.
        """
        for node in self.data.Nodes:
            node = impl(node)
            result_node = ResultNode(node, self.res1d)
            self.set_res1d_node_to_dict(result_node)
            result_node_attribute_string = make_proper_variable_name(node.ID, self.node_label)
            setattr(self, result_node_attribute_string, result_node)

    def set_res1d_node_to_dict(self, result_node):
        """
        Create a dict entry from node ID to ResultNode object.
        """
        self[result_node.id] = result_node



from .query_data import QueryData
from .query_data_catchment import QueryDataCatchment
from .query_data_global import QueryDataGlobal
from .query_data_node import QueryDataNode
from .query_data_reach import QueryDataReach
from .query_data_structure import QueryDataStructure



from ..query import QueryDataReach
from .result_location import ResultLocation


class ResultGridPoint(ResultLocation):
    """
    Class for wrapping a single ResultData grid point.

    Parameters
    ----------
    reach: IRes1DReach
        MIKE 1D IRes1DReach object.
    gridpoint IRes1DGridPoint
        MIKE 1D IRes1DGridPoint object.
    data_items : list of IDataItem objects
        A list of IDataItem objects (vector data object) the
        gridpoint has values defined on.
    res1d : Res1D
        Res1D object the grid point belongs to.

    Attributes
    ----------
    structure_data_items : list of IDataItem object.
        List of IDataItem objects belonging to a structures
        defined on the current grid point.
    """

    def __init__(self, reach, gridpoint, data_items, result_reach, res1d):
        empty_data_item_list = []
        ResultLocation.__init__(self, empty_data_item_list, res1d)
        self.reach = reach
        self.gridpoint = gridpoint
        self.result_reach = result_reach
        self.structure_data_items = []
        self.element_indices = []

    def add_to_result_quantity_maps(self, quantity_id, result_quantity):
        """Add grid point result quantity to result quantity maps."""
        self.add_to_result_quantity_map(quantity_id, result_quantity, self.result_quantity_map)

        reach_result_quantity_map = self.result_reach.result_quantity_map
        self.add_to_result_quantity_map(quantity_id, result_quantity, reach_result_quantity_map)

        reaches_result_quantity_map = self.res1d.result_network.reaches.result_quantity_map
        self.add_to_result_quantity_map(quantity_id, result_quantity, reaches_result_quantity_map)

        query = QueryDataReach(quantity_id, self.reach.Name, self.gridpoint.Chainage)
        self.add_to_network_result_quantity_map(query, result_quantity)

    def get_query(self, data_item):
        """Get a QueryDataReach for given data item."""
        quantity_id = data_item.Quantity.Id
        reach_name = self.reach.Name
        chainage = self.gridpoint.Chainage
        query = QueryDataReach(quantity_id, reach_name, chainage)
        return query

    def add_data_item(self, data_item, element_index):
        """Adds data item to grid point data items list."""
        self.data_items.append(data_item)
        self.element_indices.append(element_index)

    def add_structure_data_item(self, data_item):
        """Adds data item to structure data items list."""
        self.structure_data_items.append(data_item)



import os
import pytest
import numpy as np

from mikeio1d.custom_exceptions import NoDataForQuery, InvalidQuantity
from mikeio1d.res1d import Res1D, QueryDataCatchment
from mikeio1d.dotnet import to_numpy


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(test_folder_path, "testdata", "Swmm.out")


@pytest.fixture(params=[False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 36
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 36


@pytest.mark.parametrize(
    "query,expected",
    [
        (QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "5"), True),
        (QueryDataCatchment("SWMM_SUBCATCH_EVAP", "6"), True),
        (QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "wrong_catchment_name"), False),
    ],
)
def test_valid_catchment_data_queries(test_file, query, expected):
    swmm_out = test_file

    with pytest.raises(InvalidQuantity):
        invalid_query = QueryDataCatchment("InvalidQuantity", "20_2_2")
        assert swmm_out.read(invalid_query)

    if expected:
        swmm_out.read(query)
    else:
        with pytest.raises(NoDataForQuery):
            assert swmm_out.read(query)
        pass


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "5"), 6.562),
        (QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "6"), 1.495),
    ],
)
def test_read_reach_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


@pytest.mark.parametrize(
    "quantity,catchment_id,expected_max",
    [("SWMM_SUBCATCH_RUNOFF", "5", 6.562), ("SWMM_SUBCATCH_RUNOFF", "6", 1.495)],
)
def test_read_catchment(test_file, quantity, catchment_id, expected_max):
    data = test_file.query.GetCatchmentValues(catchment_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    print(actual_max)
    assert pytest.approx(actual_max) == expected_max


def test_time_index(test_file):
    assert len(test_file.time_index) == 36


def test_start_time(test_file):
    assert test_file.start_time == test_file.time_index.min()


def test_get_catchment_values(test_file):
    values = test_file.get_catchment_values("5", "SWMM_SUBCATCH_RUNOFF")
    assert len(values) == 36


def test_dotnet_methods(test_file):
    swmm_out = test_file
    # Just try to access the properties and methods in .net
    swmm_out.data.ResultSpecs
    swmm_out.data.Catchments
    swmm_out.query.GetCatchmentValues("5", "SWMM_SUBCATCH_RUNOFF")


def test_swmm_out_filter(test_file_path, helpers):
    catchments = ["5", "6"]
    swmm_out = Res1D(test_file_path, catchments=catchments)

    df_5 = swmm_out.read(QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "5"))
    df_6 = swmm_out.read(QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "6"))

    swmm_out_full = Res1D(test_file_path)
    df_full = swmm_out_full.read()

    helpers.assert_shared_columns_equal(df_full, df_5)
    helpers.assert_shared_columns_equal(df_full, df_6)

    # Currently Mike1D raises NullReferenceException when requesting location not included by filter
    # This should be fixed in Mike1D to raise more meaningful Mike1DException
    # with pytest.raises(Exception):
    #     assert swmm_out.read(QueryDataCatchment("SWMM_SUBCATCH_RUNOFF", "10xyz"))


def test_swmm_out_filter_readall(test_file_path):
    # Make sure readall works with filters
    catchments = ["5", "6"]
    swmm_out = Res1D(test_file_path, catchments=catchments)

    # Does not work on MIKE 1D side, giving System.NullReferenceException
    # >>> swmm_out.read()



from ..dotnet import pythonnet_implementation as impl
from .result_locations import ResultLocations
from .result_global_data import ResultGlobalData


class ResultGlobalDatas(ResultLocations):
    """
    Class for wrapping ResultData global data items.

    By itself it is also a dict, which contains
    mapping between global data item quantity ID
    and IDataItem object.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the catchments belong to.

    Attributes
    ----------
    result_global_data_list : list of ResultGlobalData objects
        list of all ResultGlobalData objects corresponding to
        all res1d.data.GlobalData.DataItems.
    """

    def __init__(self, res1d):
        ResultLocations.__init__(self, res1d)
        self.result_global_data_list = []
        self.set_global_data()

    def set_global_data(self):
        """
        Create the ResultGlobalData objects.
        No attributes are set here.
        """
        for data_item in self.data.GlobalData.DataItems:
            self.set_res1d_global_data_to_dict(data_item)
            result_global_data = ResultGlobalData(data_item, self, self.res1d)
            self.result_global_data_list.append(result_global_data)

    def set_res1d_global_data_to_dict(self, data_item):
        """
        Create a dict entry from data item quantity ID to IDatItem object.
        """
        data_item = impl(data_item)
        self[data_item.Quantity.Id] = data_item



import os
import pytest
import pandas as pd
import matplotlib as mpl

# Avoid problems with matplotlib when running on a server without a display
mpl.use("Agg")

from mikeio1d.res1d import Res1D
from mikeio1d.query import QueryDataStructure
from mikeio1d.query import QueryDataGlobal


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # Original file name was Exam6Base.res1d
    return os.path.join(test_folder_path, "testdata", "NetworkRiver.res1d")


@pytest.fixture(params=[False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 73
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 13


def test_repr(test_file):
    res1d = test_file
    res1d_repr = res1d.__repr__()
    res1d_repr_ref = (
        "<mikeio1d.Res1D>\n"
        + "Start time: 2000-02-18 00:06:00\n"
        + "End time: 2000-02-18 12:06:00\n"
        + "# Timesteps: 73\n"
        + "# Catchments: 0\n"
        + "# Nodes: 18\n"
        + "# Reaches: 18\n"
        + "# Globals: 4\n"
        + "0 - WaterLevel <m>\n"
        + "1 - Discharge <m^3/s>\n"
        + "2 - ManningResistanceNumber <m^(1/3)/s>\n"
        + "3 - FlowVelocity <m/s>\n"
        + "4 - FlowVelocityInStructure <m/s>\n"
        + "5 - FlowAreaInStructure <m^2>\n"
        + "6 - DischargeInStructure <m^3/s>\n"
        + "7 - ControlStrategyId <Integer>\n"
        + "8 - GateLevel <m>\n"
        + "9 - Variable:TwoTimeSensorGateLevel <->\n"
        + "10 - Water level:Sensor:s.h.river53745.34 <m>\n"
        + "11 - Gate level:Sensor:SensorGateLevel <m>\n"
        + "12 - Discharge:Sensor:SensorGauge1 <m^3/s>"
    )
    assert res1d_repr == res1d_repr_ref


def test_data_item_dicts(test_file):
    res1d = test_file
    assert len(res1d.catchments) == 0
    assert len(res1d.nodes) == 18
    assert len(res1d.reaches) == 9
    assert len(res1d.global_data) == 4
    assert len(res1d.structures) == 10


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataStructure("DischargeInStructure", "W_right"), 11.018),
        (QueryDataStructure("DischargeInStructure", "W_left_1_1"), 13.543),
        (QueryDataStructure("FlowAreaInStructure", "W_right"), 9.851),
        (QueryDataStructure("FlowAreaInStructure", "W_left_1_1"), 11.252),
    ],
)
def test_read_structures_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


def test_structure_attributes(test_file):
    res1d = test_file
    structures = res1d.structures

    structures.W_right.DischargeInStructure.add()

    structures.W_left_1_1.DischargeInStructure.add()

    structures.W_right.FlowAreaInStructure.add()

    structures.W_left_1_1.FlowAreaInStructure.add()

    df = res1d.read()

    actual_max = round(df["DischargeInStructure:W_right:link_basin_right:18"].max(), 3)
    assert pytest.approx(actual_max) == 11.018

    actual_max = round(df["DischargeInStructure:W_left_1_1:link_basin_left:46"].max(), 3)
    assert pytest.approx(actual_max) == 13.543

    actual_max = round(df["FlowAreaInStructure:W_right:link_basin_right:18"].max(), 3)
    assert pytest.approx(actual_max) == 9.851

    actual_max = round(df["FlowAreaInStructure:W_left_1_1:link_basin_left:46"].max(), 3)
    assert pytest.approx(actual_max) == 11.252


def test_structure_reach_attributes(test_file):
    res1d = test_file
    structures = res1d.structures

    structures.link_links_1_2_LinkChannel.Discharge.add()
    structures.link_links_2_2_LinkChannel.Discharge.add()

    df = res1d.read()

    max_discharge = round(df.max().max(), 3)
    assert pytest.approx(max_discharge) == 0.0


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataGlobal("Discharge:Sensor:SensorGauge1"), 77.522),
        (QueryDataGlobal("Gate level:Sensor:SensorGateLevel"), 53.770),
        (QueryDataGlobal("Variable:TwoTimeSensorGateLevel"), 107.54),
    ],
)
def test_read_global_items_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


def test_global_data_attributes(test_file):
    res1d = test_file
    global_data = res1d.global_data

    global_data.Discharge_Sensor_SensorGauge1.add()
    global_data.Gate_level_Sensor_SensorGateLevel.add()
    global_data.Variable_TwoTimeSensorGateLevel.add()

    df = res1d.read()

    actual_max = round(df["Discharge:Sensor:SensorGauge1"].max(), 3)
    assert pytest.approx(actual_max) == 77.522

    actual_max = round(df["Gate level:Sensor:SensorGateLevel"].max(), 3)
    assert pytest.approx(actual_max) == 53.770

    actual_max = round(df["Variable:TwoTimeSensorGateLevel"].max(), 3)
    assert pytest.approx(actual_max) == 107.54


def test_all_nodes_attributes(test_file):
    res1d = test_file
    res1d.nodes.WaterLevel.add()
    df = res1d.read()

    assert len(df.columns) == 18

    max_water_level = round(df.max().max(), 3)
    assert pytest.approx(max_water_level) == 59.200


def test_all_reach_attributes(test_file):
    res1d = test_file
    res1d.reaches.link_basin_left1_2.WaterLevel.add()
    df = res1d.read()

    assert len(df.columns) == 2

    max_water_level = round(df.max().max(), 3)
    assert pytest.approx(max_water_level) == 57.795


def test_all_reaches_attributes(test_file):
    res1d = test_file
    res1d.reaches.WaterLevel.add()
    df = res1d.read()

    assert len(df.columns) == 110

    max_water_level = round(df.max().max(), 3)
    assert pytest.approx(max_water_level) == 59.2


def test_all_structures_attributes(test_file):
    res1d = test_file
    res1d.structures.DischargeInStructure.add()
    df = res1d.read()

    assert len(df.columns) == 10

    max_discharge = round(df.max().max(), 3)
    assert pytest.approx(max_discharge) == 100.247


def test_res1d_modification(test_file):
    res1d = test_file

    # Test the reading of all the data into data frame
    df = res1d.read()
    max_value = round(df.max().max(), 3)

    # Check the overall maximum value
    assert pytest.approx(max_value) == 107.54

    # Test the modification of ResultData and
    # saving the modified data to a new res1d file.
    df2 = df.multiply(2.0)
    file_path = res1d.data.Connection.FilePath.Path
    file_path = file_path.replace("NetworkRiver.res1d", "NetworkRiver.mod.res1d")
    res1d.modify(df2, file_path=file_path)

    df_mod = res1d.read()
    max_value_mod = round(df_mod.max().max(), 3)

    # Check the overall new maximum value (should be the previous max value time two)
    assert pytest.approx(max_value_mod) == 2.0 * max_value

    # Test the reading of modified file
    res1d_new = Res1D(file_path)
    df_new = res1d_new.read()
    max_value_new = round(df_new.max().max(), 3)

    # Check the overall new maximum value again
    assert pytest.approx(max_value_new) == 2.0 * max_value


def test_res1d_modification_filtered(test_file):
    res1d = test_file

    # Test the reading of all the data into data frame
    df = res1d.read()
    max_value = round(df.max().max(), 3)

    df = df.drop(pd.date_range("2000-02-18 00:06:00", "2000-02-18 00:16:00", freq="10min"))
    df = df.drop(pd.date_range("2000-02-18 00:56:00", "2000-02-18 12:06:00", freq="10min"))

    # Test the modification of ResultData
    df2 = df.multiply(2.0)
    res1d.modify(df2)

    df_mod = res1d.read()
    max_value_mod = round(df_mod.max().max(), 3)

    # Check the overall new maximum value (should be the previous max value time two)
    assert pytest.approx(max_value_mod) == 2.0 * max_value

    # Modify the res1d data again, but only for flow velocity in structure
    res1d.structures.FlowVelocityInStructure.add()
    df = res1d.read()

    df = df.drop(pd.date_range("2000-02-18 00:06:00", "2000-02-18 00:16:00", freq="10min"))
    df = df.drop(pd.date_range("2000-02-18 00:56:00", "2000-02-18 12:06:00", freq="10min"))

    df2 = df.multiply(100.0)

    # Check the max value of flow velocity
    max_value_velocity = round(df2.max().max(), 3)
    assert pytest.approx(max_value_velocity) == 418.842

    res1d.modify(df2)
    df_mod = res1d.read()
    max_value_mod = round(df_mod.max().max(), 3)

    # Check the overall new maximum value, which should be determined by flow velocity
    assert pytest.approx(max_value_mod) == max_value_velocity


def test_extraction_to_csv_dfs0_txt(test_file):
    res1d = test_file
    res1d.clear_queries_after_reading = False

    res1d.reaches.WaterLevel.add()
    res1d.nodes.WaterLevel.add()

    file_path = res1d.data.Connection.FilePath.Path

    file_path_csv = file_path.replace("NetworkRiver.res1d", "NetworkRiver.extract.csv")
    res1d.to_csv(file_path_csv, time_step_skipping_number=10)
    file_size_csv = 21905
    assert 0.5 * file_size_csv < os.stat(file_path_csv).st_size < 2.0 * file_size_csv

    file_path_dfs0 = file_path.replace("NetworkRiver.res1d", "NetworkRiver.extract.dfs0")
    res1d.to_dfs0(file_path_dfs0, time_step_skipping_number=10)
    file_size_dfs0 = 30302
    assert file_size_dfs0 - 1000 < os.stat(file_path_dfs0).st_size < file_size_dfs0 + 1000

    file_path_txt = file_path.replace("NetworkRiver.res1d", "NetworkRiver.extract.txt")
    res1d.to_txt(file_path_txt, time_step_skipping_number=10)
    file_size_txt = 23400
    assert 0.5 * file_size_txt < os.stat(file_path_txt).st_size < 2.0 * file_size_txt


def test_result_quantity_methods(test_file):
    res1d = test_file
    file_path = res1d.data.Connection.FilePath.Path
    discharge_in_structure = res1d.structures.W_right.DischargeInStructure

    df = discharge_in_structure.to_dataframe()
    max_value = round(df.max()[0], 3)
    assert pytest.approx(max_value) == 11.018

    # Test the calling of methods
    discharge_in_structure.plot()
    discharge_in_structure.to_csv(
        file_path.replace("NetworkRiver.res1d", "W_right_DischargeInStructure.extract.csv")
    )
    discharge_in_structure.to_dfs0(
        file_path.replace("NetworkRiver.res1d", "W_right_DischargeInStructure.extract.dfs0")
    )
    discharge_in_structure.to_txt(
        file_path.replace("NetworkRiver.res1d", "W_right_DischargeInStructure.extract.txt")
    )


def test_result_quantity_collection_methods(test_file):
    res1d = test_file
    file_path = res1d.data.Connection.FilePath.Path
    discharge_in_structure = res1d.structures.DischargeInStructure

    df = discharge_in_structure.to_dataframe()
    max_value = round(df.max().max(), 3)
    assert pytest.approx(max_value) == 100.247

    # Test the calling of methods
    discharge_in_structure.plot()
    discharge_in_structure.to_csv(
        file_path.replace("NetworkRiver.res1d", "DischargeInStructure.extract.csv")
    )
    discharge_in_structure.to_dfs0(
        file_path.replace("NetworkRiver.res1d", "DischargeInStructure.extract.dfs0")
    )
    discharge_in_structure.to_txt(
        file_path.replace("NetworkRiver.res1d", "DischargeInStructure.extract.txt")
    )


def test_calculate_total_reach_lengths(res1d_river_network):
    reach = res1d_river_network.result_network.reaches.river
    assert reach._get_total_length() == pytest.approx(2024.22765)



from __future__ import annotations
import re
from typing import List, Tuple, Dict
from html import escape


class ValidPythonIdentifierTranslatorTable:
    """
    Translator table for replacing invalid python characters with a valid identifier.

    Used as argument to Python's built-in str.translate() method.

    Parameters
    ----------
    replacement : str, default = "_"
       The character to replace invalid characters with.
    """

    def __init__(self, replacement="_"):
        self.replacement = replacement

    def __getitem__(self, key):
        key = chr(key)
        if key.isdigit() or key.isidentifier():
            return key
        else:
            return self.replacement


_translator_table = ValidPythonIdentifierTranslatorTable()


def make_proper_variable_name(string: str, extra_string_before_digit="_"):
    """
    Makes a more proper variable name.

    It is assumed that the input string never is or after manipulations
    becomes an '_' or an empty string.
    """
    # Replace all characters that are not valid python identifier by an underscore.
    string = string.translate(_translator_table)
    # Replace more than two underscores with a single underscore.
    string = re.sub(r"_{2,}", "_", string)
    # Remove a starting underscore
    if len(string) > 1:
        string = string[1:] if string[0] == "_" else string
    # Remove a trailing underscore
    if len(string) > 1:
        string = string[:-1] if string[-1] == "_" else string
    # Add an extra string if the string starts with a number.
    string = extra_string_before_digit + string if string and string[0].isdigit() else string
    return string


def build_html_repr_from_sections(header: str, sections: List[Tuple[str, List | Dict]]):
    """
    Builds an html representation from a list of sections.

    Parameters
    ----------
    header : str
        Header string (e.g. <ResultNode>).
    sections : list of tuples
        List of tuples with section name and section content.
        The section content can be either a list of values or
        a dictionary of key value pairs.
    """
    repr = escape(header)
    repr += _build_html_repr_section_style()  # TODO: A better way to inject css?
    for section_name, section_content in sections:
        if isinstance(section_content, list):
            section = _build_html_repr_section_from_list(section_name, section_content)
        elif isinstance(section_content, dict):
            section = _build_html_repr_section_from_dict(section_name, section_content)
        else:
            raise ValueError(f"Unknown section content type: {section_content}")
        repr += section
    return repr


def _build_html_repr_section_style():
    """
    Builds a style string for html representation."""
    style = """
    <style>
        ul {
            margin: 0px;
            padding: 0px;
            padding-left: 2em;
        }
    </style>
    """
    return style


def _build_html_repr_section_from_dict(name, keyvalues):
    """
    Builds a section from a dictionary."""
    section = "<details>"
    section += f"<summary>{name}</summary>"
    section += "<ul>"
    for key, value in keyvalues.items():
        section += f"<li>{key}: {value}</li>"
    section += "</ul>"
    section += "</details>"
    return section


def _build_html_repr_section_from_list(name, values):
    """
    Builds a section from a list."""
    section = "<details>"
    section += f"<summary>{name}</summary>"
    section += "<ul>"
    for value in values:
        section += f"<li>{value}</li>"
    section += "</ul>"
    section += "</details>"
    return section



from ..various import NAME_DELIMITER

from .result_reader_copier import ResultReaderCopier
from .result_reader_query import ResultReaderQuery


class ResultReaderType:
    COPIER = "copier"
    QUERY = "query"


class ResultReaderCreator:
    @staticmethod
    def create(
        result_reader_type,
        res1d,
        file_path=None,
        lazy_load=False,
        header_load=False,
        reaches=None,
        nodes=None,
        catchments=None,
        col_name_delimiter=NAME_DELIMITER,
        put_chainage_in_col_name=True,
    ):
        reasult_readers = {
            ResultReaderType.COPIER: ResultReaderCopier,
            ResultReaderType.QUERY: ResultReaderQuery,
        }

        reader = reasult_readers.get(result_reader_type, None)

        return reader(
            res1d,
            file_path,
            lazy_load,
            header_load,
            reaches,
            nodes,
            catchments,
            col_name_delimiter,
            put_chainage_in_col_name,
        )



from ..various import NAME_DELIMITER
from .query_data import QueryData


class QueryDataReach(QueryData):
    """A query object that declares what reach data to extract from a .res1d file.

    Parameters
    ----------
    quantity: str
        e.g. 'WaterLevel' or 'Discharge'. Call res1d.quantities to get all quantities.
    name: str
        Reach name.
    chainage: float
        Chainage value.
    validate: bool
        Flag specifying to validate the query.

    Examples
    --------
    `QueryDataReach('WaterLevel', 'reach1', 10)` is a valid query.
    """

    def __init__(self, quantity, name=None, chainage=None, validate=True):
        super().__init__(quantity, name, validate=False)
        self._chainage = chainage

        if validate:
            self._validate()

    def _validate(self):
        super()._validate()

        if self.chainage is not None and not isinstance(self.chainage, (int, float)):
            raise TypeError("Argument 'chainage' must be either None or a number.")

        if self.name is None and self.chainage is not None:
            raise ValueError("Argument 'chainage' cannot be set if name is None.")

    def get_values(self, res1d):
        self._check_invalid_quantity(res1d)

        name = self._name
        chainage = self._chainage
        quantity = self._quantity

        values = (
            res1d.query.GetReachValues(name, chainage, quantity)
            if chainage is not None
            else res1d.query.GetReachStartValues(name, quantity)
        )

        self._check_invalid_values(values)

        return self.from_dotnet_to_python(values)

    def _update_query(self, res1d):
        name = self._name
        chainage = self._chainage
        quantity = self._quantity

        if chainage is None:
            return

        reach = res1d.searcher.FindReach(name, chainage)
        if reach is None:
            return

        data_item = res1d.query.FindDataItem(reach, quantity)
        if data_item is None:
            return

        closest_element_index = res1d.query.FindClosestElement(reach, chainage, data_item)
        if closest_element_index == -1:
            return

        gridpoint_index = list(data_item.IndexList)[closest_element_index]
        gridpoint = list(reach.GridPoints)[gridpoint_index]

        self._chainage = gridpoint.Chainage

    @property
    def chainage(self):
        return self._chainage

    def __repr__(self):
        name = self._name
        chainage = self._chainage
        quantity = self._quantity

        return (
            NAME_DELIMITER.join([quantity, name, f"{chainage:g}"])
            if chainage is not None and chainage != self.delete_value
            else NAME_DELIMITER.join([quantity, name])
        )



class Extractor:
    """Base class for data extractors to specified file format"""

    def __init__(self, out_file_name, output_data, result_data, time_step_skipping_number=1):
        self.out_file_name = out_file_name
        self.output_data = output_data
        self.result_data = result_data
        self.time_step_skipping_number = time_step_skipping_number



"""
MIKE.NET: Module for loading MIKE software .NET libraries
=========================================================

MIKE.NET is a module which creates helper methods for loading and referencing
MIKE software .NET libraries. To use the module on an existing MIKE installation
one needs before loading MIKE IO 1D to define `MIKE_INSTALL_PATH` environment
variable which points to desired installation:

```python
>>> import os
>>> os.environ["MIKE_INSTALL_PATH"] = "C:/Program File (x86)/DHI/MIKE+/2023/"
```

If this variable is not defined the MIKE IO 1D bin folder will be used per
default. This MIKE IO 1D bin folder contains very small subset of MIKE software
libraries.

Usage example
-------------

MIKE.NET can be imported as:

```python
>>> from mikeio1d import mikenet
```

When used within Jupyter `mikenet` has a list of methods starting with `load_`
which can load the relevant MIKE software libraries. For example, we can load a
`DHI.Mike1D.Generic` library as:

```python
>>> mikenet.load_Mike1D_Generic()
```

This makes a `Mike1D_Generic` reference to the `DHI.Mike1D.Generic`, which also
can be auto-completed in Jupyter. As an example, now we can create an instance
of `Quantity` class as:

```python
>>> quantity = mikenet.Mike1D_Generic.Quantity()
```

To load all MIKE software libraries one can use:

```python
>>> mikenet.load_all()
```
"""

import sys

from .library_loaders import LibraryLoaders

mikenet_module = sys.modules[__name__]
library_loaders = LibraryLoaders(mikenet_module)



import os
import pytest
import numpy as np

from mikeio1d.custom_exceptions import NoDataForQuery, InvalidQuantity
from mikeio1d.res1d import Res1D, QueryDataCatchment
from mikeio1d.dotnet import to_numpy
from mikeio1d.result_network import ResultCatchment


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # File taken from TestSuite: RainfallRunoff\SWQ\DemoSWQ1BaseMixedRRAD.res1d
    return os.path.join(test_folder_path, "testdata", "Catchments.res1d")


@pytest.fixture(params=[True, False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 108
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 5


def test_repr(test_file):
    res1d = test_file
    res1d_repr = res1d.__repr__()
    res1d_repr_ref = (
        "<mikeio1d.Res1D>\n"
        + "Start time: 1994-08-07 16:35:00\n"
        + "End time: 1994-08-07 18:35:00\n"
        + "# Timesteps: 108\n"
        + "# Catchments: 31\n"
        + "# Nodes: 0\n"
        + "# Reaches: 0\n"
        + "# Globals: 0\n"
        + "0 - TotalRunOff <m^3/s>\n"
        + "1 - ActualRainfall <m/s>\n"
        + "2 - ZinkLoadRR <kg/s>\n"
        + "3 - ZinkMassAccumulatedRR <kg>\n"
        + "4 - ZinkRR <mg/l>"
    )
    assert res1d_repr == res1d_repr_ref


def test_data_item_dicts(test_file):
    res1d = test_file
    assert len(res1d.catchments) == 31
    assert len(res1d.nodes) == 0
    assert len(res1d.reaches) == 0
    assert len(res1d.global_data) == 0


@pytest.mark.parametrize(
    "query,expected",
    [
        (QueryDataCatchment("TotalRunOff", "20_2_2"), True),
        (QueryDataCatchment("ZinkRR", "22_8_8"), True),
        (QueryDataCatchment("TotalRunOff", "wrong_catchment_name"), False),
    ],
)
def test_valid_catchment_data_queries(test_file, query, expected):
    res1d = test_file

    with pytest.raises(InvalidQuantity):
        invalid_query = QueryDataCatchment("InvalidQuantity", "20_2_2")
        assert res1d.read(invalid_query)

    if expected:
        res1d.read(query)
    else:
        with pytest.raises(NoDataForQuery):
            assert res1d.read(query)
        pass


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataCatchment("TotalRunOff", "20_2_2"), 0.236),
        (QueryDataCatchment("TotalRunOff", "22_8_8"), 0.231),
    ],
)
def test_read_catchment_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


@pytest.mark.parametrize(
    "quantity,catchment_id,expected_max",
    [("TotalRunoff", "20_2_2", 0.236), ("TotalRunoff", "22_8_8", 0.231)],
)
def test_read_catchment(test_file, quantity, catchment_id, expected_max):
    data = test_file.query.GetCatchmentValues(catchment_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    print(actual_max)
    assert pytest.approx(actual_max) == expected_max


def test_time_index(test_file):
    assert len(test_file.time_index) == 108


def test_start_time(test_file):
    assert test_file.start_time == test_file.time_index.min()


def test_get_catchment_values(test_file):
    values = test_file.get_catchment_values("20_2_2", "TotalRunOff")
    assert len(values) == 108


def test_dotnet_methods(test_file):
    res1d = test_file
    # Just try to access the properties and methods in .net
    res1d.data.ResultSpecs
    res1d.data.Catchments
    res1d.query.GetCatchmentValues("20_2_2", "TotalRunOff")


def test_res1d_filter(test_file_path, helpers):
    catchments = ["20_2_2", "22_8_8"]
    res1d = Res1D(test_file_path, catchments=catchments)

    df_20_2_2 = res1d.read(QueryDataCatchment("TotalRunOff", "20_2_2"))
    df_22_8_8 = res1d.read(QueryDataCatchment("TotalRunOff", "22_8_8"))

    res1d_full = Res1D(test_file_path)
    df_full = res1d_full.read()

    helpers.assert_shared_columns_equal(df_full, df_20_2_2)
    helpers.assert_shared_columns_equal(df_full, df_22_8_8)

    # Currently Mike1D raises NullReferenceException when requesting location not included by filter
    # This should be fixed in Mike1D to raise more meaningful Mike1DException
    # with pytest.raises(Exception):
    #     assert res1d.read(QueryDataCatchment("TotalRunOff", "100_16_16"))


def test_res1d_filter_readall(test_file_path, helpers):
    # Make sure readall works with filters
    catchments = ["20_2_2", "22_8_8"]
    res1d = Res1D(test_file_path, catchments=catchments)
    df = res1d.read()

    res1d_full = Res1D(test_file_path)
    df_full = res1d_full.read()

    helpers.assert_shared_columns_equal(df_full, df)


def test_catchment_attributes(test_file):
    res1d = test_file
    catchments = res1d.catchments

    catchments.c_20_2_2.TotalRunOff.add()
    catchments.c_22_8_8.TotalRunOff.add()
    df = res1d.read()

    actual_max = round(df["TotalRunOff:20_2_2"].max(), 3)
    assert pytest.approx(actual_max) == 0.236

    actual_max = round(df["TotalRunOff:22_8_8"].max(), 3)
    assert pytest.approx(actual_max) == 0.231


def test_all_catchments_attributes(test_file):
    res1d = test_file
    res1d.catchments.TotalRunOff.add()
    df = res1d.read()

    assert len(df.columns) == 31

    max_runoff = round(df.max().max(), 3)
    assert pytest.approx(max_runoff) == 0.469


def test_catchment_attribute_access_maintains_backwards_compat(res1d_catchments):
    res = res1d_catchments
    catchment = res.result_network.catchments["100_16_16"]
    with pytest.warns(UserWarning):
        assert catchment.Id == "100_16_16"
        assert pytest.approx(catchment.Area) == 22800.0
    for name, catchment in res.result_network.catchments.items():
        assert catchment.CatchmentName == name


def test_catchments_map_to_python_catchment(res1d_catchments):
    res = res1d_catchments
    catchment = res.result_network.catchments["100_16_16"]
    assert isinstance(catchment, ResultCatchment)


def test_catchment_static_attributes(res1d_catchments):
    res = res1d_catchments
    catchment = res.result_network.catchments["100_16_16"]
    assert catchment.id == "100_16_16"
    assert pytest.approx(catchment.area) == 22800.0
    assert catchment.type == "Kinematic Wave"
    # check that static attributes are accessible for all catchments
    for catchment in res.result_network.catchments.values():
        catchment.id
        catchment.area
        catchment.type



from .extractor import Extractor

import System
from System import Array

from DHI.Generic.MikeZero import eumUnit
from DHI.Generic.MikeZero.DFS import DfsFactory
from DHI.Generic.MikeZero.DFS import DfsBuilder
from DHI.Generic.MikeZero.DFS import DfsSimpleType
from DHI.Generic.MikeZero.DFS import DataValueType
from DHI.Generic.MikeZero.DFS import StatType

from DHI.Mike1D.ResultDataAccess import ItemTypeGroup


class ExtractorDfs0(Extractor):
    """Class which extracts data to dfs0 file format"""

    def export(self):
        self.factory = DfsFactory()
        self.builder = self.create_dfs_builder()
        self.define_dynamic_data_items()
        self.write_data_items()

    def create_dfs_builder(self):
        result_data = self.result_data
        factory = self.factory

        builder = DfsBuilder.Create("mikeio1d", "MIKE SDK", 100)

        # Set up file header
        builder.SetDataType(1)
        builder.SetGeographicalProjection(factory.CreateProjectionUndefined())
        builder.SetTemporalAxis(
            factory.CreateTemporalNonEqCalendarAxis(eumUnit.eumUsec, result_data.StartTime)
        )
        builder.SetItemStatisticsType(StatType.NoStat)

        return builder

    def define_dynamic_data_items(self):
        output_data = self.output_data
        result_data = self.result_data
        builder = self.builder

        for data_entry in output_data:
            data_item = data_entry.data_item
            element_index = data_entry.element_index

            quantity = data_item.Quantity
            item_type_group = data_item.ItemTypeGroup
            number_within_group = data_item.NumberWithinGroup

            reaches = list(result_data.Reaches)
            nodes = list(result_data.Nodes)
            catchments = list(result_data.Catchments)

            if item_type_group == ItemTypeGroup.ReachItem:
                reach = reaches[number_within_group]
                gridpoint_index = data_item.IndexList[element_index]
                gridpoints = list(reach.GridPoints)
                chainage = gridpoints[gridpoint_index].Chainage
                item_name = "reach:%s:%s:%.3f" % (quantity.Id, reach.Name, chainage)

            elif item_type_group == ItemTypeGroup.NodeItem:
                node = nodes[number_within_group]
                item_name = "node:%s:%s" % (quantity.Id, node.Id)

            elif item_type_group == ItemTypeGroup.CatchmentItem:
                catchment = catchments[number_within_group]
                item_name = "catchment:%s:%s" % (quantity.Id, catchment.Id)

            else:
                item_name = "%s:%s:%s" % (item_type_group, quantity.Id, data_item.Id)

            item = builder.CreateDynamicItemBuilder()
            item.Set(item_name, data_item.Quantity.EumQuantity, DfsSimpleType.Float)
            item.SetValueType(DataValueType.Instantaneous)
            item.SetAxis(self.factory.CreateAxisEqD0())
            builder.AddDynamicItem(item.GetDynamicItemInfo())

    def write_data_items(self):
        output_data = self.output_data
        result_data = self.result_data
        builder = self.builder

        # Create file
        builder.CreateFile(self.out_file_name)
        dfsfile = builder.GetFile()
        times = list(result_data.TimesList)

        # Write data to file
        val = Array.CreateInstance(System.Single, 1)
        for time_step_index in range(result_data.NumberOfTimeSteps):
            if time_step_index % self.time_step_skipping_number != 0:
                continue

            time = times[time_step_index].Subtract(result_data.StartTime).TotalSeconds
            for data_entry in output_data:
                data_item = data_entry.data_item
                element_index = data_entry.element_index

                val[0] = data_item.GetValue(time_step_index, element_index)
                dfsfile.WriteItemTimeStepNext(time, val)

        dfsfile.Close()



import os


class MikePath:
    """
    Class for handling MIKE binary paths.

    Attributes
    ----------
    mikeio1d_bin_path: string
        Path to binaries contained in MIKE IO 1D Python package.
    mike_install_path: string
        Environment variable MIKE_INSTALL_PATH specifying location of MIKE installation.
    skip_bin_x64: string
        Environment variable specifying not to append 'bin/x64' for binary location.
    bin_x64: string
        String which will be appended to mike_install_path for full location of binaries.
    mike_bin_path: string
        Actual path for MIKE binaries, which is either mikeio1d_bin_path or mike_install_path/bin/x64.
    """

    mikeio1d_bin_path = os.path.join(os.path.dirname(__file__), "bin")

    mike_install_path = os.environ.get("MIKE_INSTALL_PATH", None)

    skip_bin_x64 = os.environ.get("MIKE_SKIP_BIN_X64", None)
    bin_x64 = os.path.join("bin", "x64") if skip_bin_x64 is None else ""

    mike_bin_path = (
        os.path.join(mike_install_path, bin_x64)
        if mike_install_path is not None
        else mikeio1d_bin_path
    )

    library_patterns = ["DHI.*.dll"]

    @staticmethod
    def setup_mike_installation(syspath):
        """
        Adds MIKE installation bin path and MIKE IO 1D bin path to the
        given PATH variable.

        Parameters
        ----------
        syspath: list of str
            List of strings defining PATH variable.
        """
        mike_bin_path = MikePath.mike_bin_path
        mikeio1d_bin_path = MikePath.mikeio1d_bin_path
        mike_install_path = MikePath.mike_install_path

        syspath.append(mike_bin_path)
        if mikeio1d_bin_path != mike_bin_path:
            syspath.append(mikeio1d_bin_path)

            # Some of the MIKE libraries will need to be resolved by DHI.Mike.Install,
            # so set it up here.
            import clr

            clr.AddReference(
                "DHI.Mike.Install, Version=1.0.0.0, Culture=neutral, PublicKeyToken=c513450b5d0bf0bf"
            )
            from DHI.Mike.Install import MikeImport

            MikeImport.SetupInstallDir(mike_install_path)



import dataclasses
from dataclasses import dataclass
from pathlib import Path


def format_path(filename):
    path = Path(__file__).parent / filename
    return str(path.absolute())


# Using dataclass worked better for autocompletion in vscode
@dataclass
class testdata:
    """
    A class that contains file paths for test data used in the mikeio1d package.
    Useful for testing and examples, without having to get path. Works for unsaved
    notebooks as well.

    Examples
    --------
    >>> from mikeio1d import Res1D
    >>> from tests import testdata
    >>> res = Res1D(testdata.Network_res1d)
    """

    Network_res1d: str = format_path("Network.res1d")
    """A basic urban network file."""
    Network_res1d_chinese: str = format_path("Network_chinese.res1d")
    """A basic urban network file with chinese characters for some links."""
    NetworkRiver_res1d: str = format_path("NetworkRiver.res1d")
    """A basic river network file."""
    Catchments_res1d: str = format_path("Catchments.res1d")
    """A small urban setup with three pipes."""
    FlowSplit_res1d: str = format_path("FlowSplit.res1d")
    """A basic urban network file containing only catchments."""
    LTSEventStatistics_res1d: str = format_path("LTSEventStatistics.res1d")
    """An LTS event statistics file."""
    LTSMonthlyStatistics_res1d: str = format_path("LTSMonthlyStatistics.res1d")
    """An LTS monthly statistics file."""
    xsections_xns11: str = format_path("xsections.xns11")
    """A basic xsections file."""
    Epanet_res: str = format_path("Epanet.res")
    """A basic Epanet res file. Must have accompanying .inp file."""
    Epanet_resx: str = format_path("Epanet.resx")
    """A basic Epanet resx file. Must have accompanying .inp file."""
    SWMM_out: str = format_path("SWMM.out")
    """A basic SWMM result file. Must have accompanying .inp file."""

    def get_expected_dataframe(self, name: str):
        import pandas as pd

        return pd.read_parquet(Path(__file__).parent / "expected_results" / f"{name}.parquet")


testdata = testdata()


def generate_expected_results(output_folder: Path = Path(__file__).parent / "expected_results"):
    import pandas as pd
    from pandas.testing import assert_frame_equal
    from mikeio1d import Res1D

    if not output_folder.exists():
        output_folder.mkdir()
    for name, path in dataclasses.asdict(testdata).items():
        if path.endswith(".xns11"):
            continue
        res = Res1D(path)
        df = res.read()
        df = df.loc[
            :, ~df.columns.duplicated()
        ]  # TODO: Remove this when column names are guaranteed unique
        output_path = output_folder / f"{name}.parquet"
        df.to_parquet(output_path)
        df_parquet = testdata.get_expected_dataframe(name)
        assert_frame_equal(df, df_parquet)



import clr
from pathlib import Path
from warnings import warn


class LibraryLoader:
    """
    Helper class which assigns 'load_[LIRABRY_NAME]' to MIKE.NET module.

    Parameters
    ----------
    file_name: str
        Library file_name to load. Typically this is a full path to the library (dll).
    mikenet_module: module
        Reference to MIKET.NET module
    """

    def __init__(self, file_name, mikenet_module):
        self.file_name = file_name
        self.mikenet_module = mikenet_module

        self.file_path = Path(file_name)
        self.library_name = self.file_path.stem
        self.library_alias = self.library_name.replace("DHI.", "")
        self.library_alias = self.library_alias.replace(".", "_")

        load_function_name = "load_" + self.library_alias

        setattr(mikenet_module, load_function_name, self.load)

    def load(self):
        """
        A method which adds a reference of a relevant library to Python.NET (CLR)
        and also adds a reference to the MIKE.NET.
        """
        clr.AddReference(self.library_name)

        try:
            mikenet_dict = self.mikenet_module.__dict__
            exec(f"import {self.library_name} as {self.library_alias}", mikenet_dict)
        except Exception as e:
            warn(f"Could not import .NET library {self.library_name}: {e}")



import pytest
from .testdata import testdata

from pandas.testing import assert_frame_equal

from mikeio1d import Res1D


def testdata_name():
    import dataclasses

    return list(dataclasses.asdict(testdata).keys())


@pytest.mark.parametrize("extension", [".res1d", ".res", ".resx", ".out"])
def test_mikeio1d_generates_expected_dataframe_for_filetype(extension):
    for name in testdata_name():
        path = getattr(testdata, name)
        if not path.endswith(extension):
            continue
        df = Res1D(path).read()
        df = df.loc[
            :, ~df.columns.duplicated()
        ]  # TODO: Remove this when column names are guaranteed unique
        df_expected = testdata.get_expected_dataframe(name)
        assert_frame_equal(df, df_expected)



import os
import pytest
import numpy as np

from mikeio1d.custom_exceptions import NoDataForQuery, InvalidQuantity
from mikeio1d.res1d import Res1D, QueryDataReach, QueryDataNode
from mikeio1d.dotnet import to_numpy


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # Original file name was Exam6Base.epanet_resx
    return os.path.join(test_folder_path, "testdata", "Epanet.resx")


@pytest.fixture(params=[False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 25
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 5


def test_repr(test_file):
    epanet_resx = test_file
    epanet_resx_repr = epanet_resx.__repr__()
    epanet_repr_ref = (
        "<mikeio1d.Res1D>\n"
        + "Start time: 2022-10-13 00:00:00\n"
        + "End time: 2022-10-14 00:00:00\n"
        + "# Timesteps: 25\n"
        + "# Catchments: 0\n"
        + "# Nodes: 2\n"
        + "# Reaches: 1\n"
        + "# Globals: 0\n"
        + "0 - Volume <m^3>\n"
        + "1 - Volume Percentage <%>\n"
        + "2 - Pump efficiency <%>\n"
        + "3 - Pump energy costs </kWh>\n"
        + "4 - Pump energy <kW>"
    )
    assert epanet_resx_repr == epanet_repr_ref


def test_data_item_dicts(test_file):
    epanet_resx = test_file
    assert len(epanet_resx.catchments) == 0
    assert len(epanet_resx.nodes) == 2
    assert len(epanet_resx.reaches) == 1
    assert len(epanet_resx.global_data) == 0


@pytest.mark.parametrize(
    "query,expected",
    [
        (QueryDataReach("Pump energy", "9"), True),
        (QueryDataReach("Pump energy", "10xyz"), False),
        (QueryDataReach("Pump energy", "wrong_reach_name"), False),
    ],
)
def test_valid_reach_data_queries(test_file, query, expected):
    epanet_resx = test_file

    with pytest.raises(InvalidQuantity):
        invalid_query = QueryDataReach("InvalidQuantity", "9")
        assert epanet_resx.read(invalid_query)

    if expected:
        epanet_resx.read(query)
    else:
        with pytest.raises(Exception):
            assert epanet_resx.read(query)


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataReach("Pump energy", "9"), 96.707),
        (QueryDataReach("Pump energy costs", "9"), 0.0),
        (QueryDataReach("Pump efficiency", "9"), 75.0),
    ],
)
def test_read_reach_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


@pytest.mark.parametrize(
    "quantity,reach_id,expected_max",
    [("Pump energy", "9", 96.707), ("Pump energy costs", "9", 0.0), ("Pump efficiency", "9", 75.0)],
)
def test_read_reach(test_file, quantity, reach_id, expected_max):
    data = test_file.query.GetReachStartValues(reach_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


@pytest.mark.parametrize(
    "quantity,node_id,expected_max", [("Volume", "2", 7859.459), ("Volume Percentage", "2", 92.381)]
)
def test_read_node(test_file, quantity, node_id, expected_max):
    data = test_file.query.GetNodeValues(node_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


def test_time_index(test_file):
    assert len(test_file.time_index) == 25


def test_start_time(test_file):
    assert test_file.start_time == test_file.time_index.min()


def test_get_node_values(test_file):
    values = test_file.get_node_values("9", "Volume")
    assert len(values) == 25


def test_get_reach_values(test_file):
    # When reading EPANET results there is a bug in MIKE 1D,
    # which does not allow to call GetReachValues. This does not work
    # >>> values = test_file.get_reach_values("9", 0, "Pump energy")
    # >>> time_series = pd.Series(values, index=test_file.time_index)
    # >>> assert len(values) == 25
    # >>> assert len(time_series.index) == 25

    # Just try to call the methods
    test_file.get_reach_end_values("9", "Pump energy")
    test_file.get_reach_start_values("9", "Pump energy")
    test_file.get_reach_sum_values("9", "Pump energy")


def test_get_reach_value(test_file):
    # Does not work from MIKE 1D side.
    # >>> value = test_file.get_reach_value("9", 0, "Pump energy", test_file.start_time)
    # >>> assert value > 0
    assert 0 == 0


def test_dotnet_methods(test_file):
    epanet_resx = test_file
    epanet_resx.data.ResultSpecs
    epanet_resx.data.Nodes

    assert pytest.approx(6806.107) == epanet_resx.query.GetNodeValues("2", "Volume")[0]

    # Does not work from MIKE 1D side.
    # >>> epanet_resx.query.GetReachValue("9", 0, "Pump energy", epanet_resx.data.StartTime)
    # >>> epanet_resx.query.GetReachValues("9", 0, "Pump energy")

    assert pytest.approx(95.8442) == epanet_resx.query.GetReachEndValues("9", "Pump energy")[0]
    assert pytest.approx(95.8442) == epanet_resx.query.GetReachStartValues("9", "Pump energy")[0]
    assert pytest.approx(95.8442) == epanet_resx.query.GetReachSumValues("9", "Pump energy")[0]


def test_epanet_resx_filter(test_file_path, helpers):
    nodes = ["2", "9"]
    reaches = ["9"]
    epanet_resx = Res1D(test_file_path, nodes=nodes, reaches=reaches)

    df_energy_9 = epanet_resx.read(QueryDataReach("Pump energy", "9"))
    df_volume_2 = epanet_resx.read(QueryDataNode("Volume", "2"))
    df_volume_9 = epanet_resx.read(QueryDataNode("Volume", "9"))

    epanet_resx_full = Res1D(test_file_path)
    df_full = epanet_resx_full.read()

    helpers.assert_shared_columns_equal(df_full, df_energy_9)
    helpers.assert_shared_columns_equal(df_full, df_volume_2)
    helpers.assert_shared_columns_equal(df_full, df_volume_9)

    # Currently Mike1D raises System.ArgumentOutOfRangeException when requesting location not included by filter
    # This should be fixed in Mike1D to raise more meaningful Mike1DException
    with pytest.raises(Exception):
        assert epanet_resx.read(QueryDataReach("Pump energy", "10xyz"))

    with pytest.raises(NoDataForQuery):
        assert epanet_resx.read(QueryDataNode("Volume", "10xyz"))


def test_epanet_resx_filter_readall(test_file_path, helpers):
    # Make sure read all can be used with filters
    nodes = ["2", "9"]
    reaches = ["9"]
    epanet_resx = Res1D(test_file_path, nodes=nodes, reaches=reaches)
    df = epanet_resx.read()

    epanet_resx_full = Res1D(test_file_path)
    df_full = epanet_resx_full.read()

    helpers.assert_shared_columns_equal(df_full, df)



from warnings import warn

from ..dotnet import pythonnet_implementation as impl
from ..query import QueryDataStructure
from .result_location import ResultLocation


class ResultStructure(ResultLocation):
    """
    Class for wrapping a list of ResultData structure data items
    belonging to the same structure.

    Parameters
    ----------
    structure_id : str
        Structure ID.
    reach: IRes1DReach
        Reach where the structure belongs to.
    res1d : Res1D
        Res1D object the reach belongs to.

    Attributes
    ----------
    data_items : list of IDataItems objects.
        A list of MIKE 1D IDataItems objects corresponding to a given structure.
    data_items_dict : dict
        A dictionary from quantity id to a data item.
    chainage : float
        Chainage where the structure is located on the reach.
    """

    def __init__(self, structure_id, reach, data_items, res1d):
        empty_list = []
        ResultLocation.__init__(self, empty_list, res1d)

        self.id = structure_id
        self.reach = reach
        self.chainage = None

        self.data_items_dict = {}
        for data_item in data_items:
            self.add_res1d_structure_data_item(data_item)

        self.set_static_attributes()

    def __repr__(self) -> str:
        return f"<{self.type}: {self.id}>"

    @property
    def structure_id(self):
        # TODO: Remove this in 1.0.0
        warn(
            "Please use .id instead of .structure_id. This attribute will be removed in the future."
        )
        return self.id

    def set_static_attributes(self):
        """Set static attributes. These show up in the html repr."""
        self._static_attributes = []
        self.set_static_attribute("id", self.id)
        self.set_static_attribute("type", self.reach.Name.split(":")[0])
        self.set_static_attribute("chainage", self.chainage)

    def add_to_result_quantity_maps(self, quantity_id, result_quantity):
        """Add structure result quantity to result quantity maps."""
        self.add_to_result_quantity_map(quantity_id, result_quantity, self.result_quantity_map)

        structure_result_quantity_map = self.res1d.result_network.structures.result_quantity_map
        self.add_to_result_quantity_map(quantity_id, result_quantity, structure_result_quantity_map)

        query = QueryDataStructure(quantity_id, self.id, self.reach.Name, self.chainage)
        self.add_to_network_result_quantity_map(query, result_quantity)

    def add_res1d_structure_data_item(self, data_item):
        """
        Add a IDataItem to ResultStructure.

        Parameters
        ----------
        data_item: IDataItem
            A MIKE 1D IDataItem object.
        """
        if self.chainage is None:
            index_list = list(data_item.IndexList)
            gridpoint_index = index_list[0]
            gridpoints = list(self.reach.GridPoints)
            self.chainage = gridpoints[gridpoint_index].Chainage

        self.data_items.append(data_item)
        self.data_items_dict[data_item.Quantity.Id] = data_item
        self.set_quantity(self, data_item)

    @staticmethod
    def get_structure_id(reach, data_item):
        """
        Gets structure ID either from IDataItem.ItemId or for structure reaches
        from actual Res1DStructureGridPoint structure.
        """
        if data_item.ItemId is not None:
            return data_item.ItemId

        if reach.IsStructureReach:
            structure_gridpoint = impl(list(reach.GridPoints)[1])
            structures = list(structure_gridpoint.Structures)
            structure_id = structures[0].Id
            return structure_id

        return None

    def get_data_item(self, quantity_id):
        """Retrieve a data item for given quantity id."""
        return self.data_items_dict[quantity_id]

    def get_query(self, data_item):
        """Get a QueryDataStructure for given data item."""
        quantity_id = data_item.Quantity.Id
        structure_id = self.id
        query = QueryDataStructure(quantity_id, structure_id, self.reach.Name, self.chainage)
        return query



from ..query import QueryDataGlobal
from .result_location import ResultLocation


class ResultGlobalData(ResultLocation):
    """
    Class for wrapping ResultData global data items.

    By itself it is also a dict, which contains
    mapping between data item quantity ID and IDataItem object.

    Parameters
    ----------
    data_item : IDataItem
        MIKE 1D IDataItem object for a global data item.
    global_datas : ResultGlobalDatas object.
        A wrapper object for all global data items.
    res1d : Res1D
        Res1D object the global data belongs to.
    """

    def __init__(self, data_item, global_datas, res1d):
        ResultLocation.__init__(self, [data_item], res1d)
        self.data_item = data_item
        self.global_datas = global_datas
        self.set_quantities()

    def set_quantities(self):
        """
        Override of base set_quantities.
        Here only a single data item is used for ResultGlobalData.
        Also the quantity attribute is assigned to self.global_data.
        """
        data_item = self.data_item
        self.set_quantity(self.global_datas, data_item)

    def add_to_result_quantity_maps(self, quantity_id, result_quantity):
        """Add global data result quantity to result quantity maps."""
        query = QueryDataGlobal(quantity_id)
        self.add_to_network_result_quantity_map(query, result_quantity)

    def get_query(self, data_item):
        """Get a QueryDataGlobal for given data item."""
        quantity_id = data_item.Quantity.Id
        query = QueryDataGlobal(quantity_id)
        return query



from .result_query import QueryData
from .result_query import QueryDataCatchment
from .result_query import QueryDataGlobal
from .result_query import QueryDataNode
from .result_query import QueryDataReach
from .result_query import QueryDataStructure



from .extractor_dfs0 import ExtractorDfs0
from .extractor_csv import ExtractorCsv
from .extractor_txt import ExtractorTxt


class ExtractorAll:
    """Class which extracts data into all supported file formats"""

    def __init__(self, out_file_name, output_data, result_data, time_step_skipping_number=1):
        self.all_extractors = [
            ExtractorTxt(
                out_file_name.replace(".-", ".txt"),
                output_data,
                result_data,
                time_step_skipping_number,
            ),
            ExtractorCsv(
                out_file_name.replace(".-", ".csv"),
                output_data,
                result_data,
                time_step_skipping_number,
            ),
            ExtractorDfs0(
                out_file_name.replace(".-", ".dfs0"),
                output_data,
                result_data,
                time_step_skipping_number,
            ),
        ]

    def export(self):
        for extractor in self.all_extractors:
            extractor.export()



from .extractor_all import ExtractorAll
from .extractor_dfs0 import ExtractorDfs0
from .extractor_csv import ExtractorCsv
from .extractor_txt import ExtractorTxt


class ExtractorOutputFileType:
    TXT = "txt"
    CSV = "csv"
    DFS0 = "dfs0"
    ALL = "-"


class ExtractorCreator:
    """Class which creates an ResultData time series extractors"""

    @staticmethod
    def create(out_file_type, out_file_name, output_data, result_data, time_step_skipping_number=1):
        extractors = {
            ExtractorOutputFileType.TXT: ExtractorTxt,
            ExtractorOutputFileType.CSV: ExtractorCsv,
            ExtractorOutputFileType.DFS0: ExtractorDfs0,
            ExtractorOutputFileType.ALL: ExtractorAll,
        }

        extractor = extractors.get(out_file_type, None)

        return extractor(out_file_name, output_data, result_data, time_step_skipping_number)



import numpy as np
import pandas as pd

from ..dotnet import pythonnet_implementation as impl
from ..various import NAME_DELIMITER
from .result_reader import ResultReader

from System import IntPtr

from DHI.Mike1D.MikeIO import ResultDataCopier
from DHI.Mike1D.MikeIO import DataEntry as DataEntryNet


class ResultReaderCopier(ResultReader):
    """
    Class for reading the ResultData object TimeData
    into Pandas data frame using ResultDataCopier object
    from DHI.Mike1D.MikeIO library.
    """

    def __init__(
        self,
        res1d,
        file_path=None,
        lazy_load=False,
        header_load=False,
        reaches=None,
        nodes=None,
        catchments=None,
        col_name_delimiter=NAME_DELIMITER,
        put_chainage_in_col_name=True,
    ):
        ResultReader.__init__(
            self,
            res1d,
            file_path,
            lazy_load,
            header_load,
            reaches,
            nodes,
            catchments,
            col_name_delimiter,
            put_chainage_in_col_name,
        )

        self.result_data_copier = ResultDataCopier(self.data)

    def read(self, queries=None):
        column_names = []
        data_entries = self.result_data_copier.GetEmptyDataEntriesList()
        for query in queries:
            query.add_to_data_entries(self.res1d, data_entries, column_names)

        df = self.create_data_frame(data_entries, column_names)
        return df

    def read_all(self):
        data_entries, column_names = self.get_all_data_entries_and_column_names()
        df = self.create_data_frame(data_entries, column_names)
        return df

    def create_data_frame(self, data_entries, column_names):
        number_of_timesteps = self.data.NumberOfTimeSteps
        number_of_items = len(data_entries)

        shape = (number_of_timesteps, number_of_items)
        data_array = np.zeros(shape, dtype=np.dtype("float32"), order="F")

        data_pointer = data_array.ctypes.data
        data_pointer_net = IntPtr(data_pointer)
        self.result_data_copier.CopyData(data_pointer_net, data_entries)

        df = pd.DataFrame(data_array, index=self.time_index, columns=column_names)

        self.update_time_quantities(df)

        return df

    def get_all_data_entries_and_column_names(self):
        data_entries = self.result_data_copier.GetEmptyDataEntriesList()
        column_names = []
        for data_set in self.data.DataSets:
            data_set = impl(data_set)
            if not self.is_data_set_included(data_set):
                continue

            for data_item in data_set.DataItems:
                for i in range(data_item.NumberOfElements):
                    data_entry = DataEntryNet(data_item, i)
                    data_entries.Add(data_entry)

                    column_name = self.get_column_name(data_set, data_item, i)
                    column_names.append(column_name)

        return data_entries, column_names



"""
Tests for the formatting of the result network objects."""


def test_location_str_repr_header_line(node, reach, catchment, structure):
    assert node.__repr__().startswith("<Manhole: 1>"), "Node header line is not correct."
    assert reach.__repr__().startswith("<Reach: 100l1>"), "Reach header line is not correct."
    assert catchment.__repr__().startswith(
        "<Catchment: 100_16_16>"
    ), "Catchment header line is not correct."
    assert structure.__repr__().startswith("<Weir: 119w1>"), "Structure header line is not correct."


def test_location_html_repr_header_line(node, reach, catchment):
    assert node._repr_html_().startswith("&lt;Manhole: 1&gt;"), "Node header line is not correct."
    assert reach._repr_html_().startswith(
        "&lt;Reach: 100l1&gt;"
    ), "Reach header line is not correct."
    assert catchment._repr_html_().startswith(
        "&lt;Catchment: 100_16_16&gt;"
    ), "Catchment header line is not correct."


def test_node_html_repr_runs(many_nodes):
    for node in many_nodes:
        node._repr_html_()


def test_reach_html_repr_runs(many_reaches):
    for reach in many_reaches:
        reach._repr_html_()


def test_catchment_html_repr_runs(many_catchments):
    for catchment in many_catchments:
        catchment._repr_html_()


def test_single_node_html_repr(node):
    html_repr = node._repr_html_()
    expected_html_repr = """&lt;Manhole: 1&gt;
    <style>
        ul {
            margin: 0px;
            padding: 0px;
            padding-left: 2em;
        }
    </style>
    <details><summary>Attributes</summary><ul><li>id: 1</li><li>type: Manhole</li><li>xcoord: -687934.6000976562</li><li>ycoord: -1056500.69921875</li><li>ground_level: 197.07000732421875</li><li>bottom_level: 195.0500030517578</li><li>critical_level: inf</li><li>diameter: 1.0</li></ul></details><details><summary>Quantities</summary><ul><li>WaterLevel</li></ul></details>"""
    assert html_repr == expected_html_repr


def test_single_catchment_html_repr(catchment):
    html_repr = catchment._repr_html_()
    expected_html_repr = """&lt;Catchment: 100_16_16&gt;
    <style>
        ul {
            margin: 0px;
            padding: 0px;
            padding-left: 2em;
        }
    </style>
    <details><summary>Attributes</summary><ul><li>id: 100_16_16</li><li>area: 22800.0</li><li>type: Kinematic Wave</li></ul></details><details><summary>Quantities</summary><ul><li>TotalRunOff</li><li>ActualRainfall</li><li>ZinkLoadRR</li><li>ZinkMassAccumulatedRR</li><li>ZinkRR</li></ul></details>"""
    assert html_repr == expected_html_repr


def test_single_reach_html_repr(river_reach):
    html_repr = river_reach._repr_html_()
    excepted_html_repr = """&lt;Reach: river&gt;
    <style>
        ul {
            margin: 0px;
            padding: 0px;
            padding-left: 2em;
        }
    </style>
    <details><summary>Attributes</summary><ul><li>name: river</li><li>length: 2024.2276598819008</li><li>start_chainage: 53100.0</li><li>end_chainage: 55124.2276598819</li><li>n_gridpoints: 6</li></ul></details><details><summary>Quantities</summary><ul><li>WaterLevel</li><li>ManningResistanceNumber</li><li>Discharge</li><li>FlowVelocity</li></ul></details>"""
    assert html_repr == excepted_html_repr


def test_single_structure_html_repr(structure):
    html_repr = structure._repr_html_()
    expected_html_repr = """&lt;Weir: 119w1&gt;
    <style>
        ul {
            margin: 0px;
            padding: 0px;
            padding-left: 2em;
        }
    </style>
    <details><summary>Attributes</summary><ul><li>id: 119w1</li><li>type: Weir</li><li>chainage: 0.5</li></ul></details><details><summary>Quantities</summary><ul><li>Discharge</li></ul></details>"""
    assert html_repr == expected_html_repr



import clr
import sys
import os
from platform import architecture

from .mikepath import MikePath

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#
__version__ = "0.4.1"

if "64" not in architecture()[0]:
    raise Exception("This library has not been tested for a 32 bit system.")

MikePath.setup_mike_installation(sys.path)

clr.AddReference("System")
clr.AddReference("System.Runtime")
clr.AddReference("System.Runtime.InteropServices")
clr.AddReference("DHI.Generic.MikeZero.DFS")
clr.AddReference("DHI.Generic.MikeZero.EUM")
# clr.AddReference('DHI.PFS')
# clr.AddReference('DHI.Projections')
clr.AddReference("DHI.Mike1D.Generic")
clr.AddReference("DHI.Mike1D.ResultDataAccess")
clr.AddReference("DHI.Mike1D.CrossSectionModule")
clr.AddReference("DHI.Mike1D.MikeIO")

from .res1d import Res1D
